use anyhow::Result;
use serde::Serialize;
use std::path::Path;
use std::sync::{Arc, Mutex};

use crate::db::Database;

const CHANGELOG_MD: &str = include_str!("../CHANGELOG.md");

pub fn changelog() -> String {
    comrak::markdown_to_html(CHANGELOG_MD, &comrak::ComrakOptions::default())
}

#[derive(PartialEq, Serialize, Debug)]
pub enum ImportStatus {
    Pending,
    Loading(f32),
    Finished,
    Failed(String),
}
impl std::fmt::Display for ImportStatus {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::Pending => write!(f, "queued"),
            Self::Loading(p) => write!(f, "{:.0}%", p * 100.0),
            Self::Finished => write!(f, "imported"),
            Self::Failed(e) => write!(f, "failed ({})", e),
        }
    }
}

#[derive(Serialize, Debug)]
pub struct ImportQueueEntry {
    pub path: String,
    pub status: Mutex<ImportStatus>,
    pub delete_after: bool,
}

pub type ImportQueueEntryRef = Arc<ImportQueueEntry>;

#[derive(Default, Clone)]
pub struct ImportQueue(Arc<Mutex<Vec<ImportQueueEntryRef>>>);
impl ImportQueue {
    pub fn push(&self, path: String, delete_after: bool) -> Result<()> {
        let entry = Arc::new(ImportQueueEntry {
            path,
            status: Mutex::new(ImportStatus::Pending),
            delete_after,
        });
        self.0.lock().expect("failed to lock queue").push(entry);
        Ok(())
    }
    pub fn status(&self) -> Vec<ImportQueueEntryRef> {
        self.0
            .lock()
            .expect("failed to lock queue")
            .iter()
            .cloned()
            .collect()
    }
    pub fn next(&self) -> Option<ImportQueueEntryRef> {
        self.0
            .lock()
            .expect("failed to lock queue")
            .iter()
            .find(|e| *(e.status.lock().unwrap()) == ImportStatus::Pending)
            .cloned()
    }
    pub fn clear(&self) {
        self.0.lock().expect("failed to lock queue").clear()
    }
}

#[derive(Default)]
pub struct RuntimeSettings {
    pub main_url: Option<String>,
    /// set when the side web server for the gui settles on an endpoint to listen on
    pub attachments_url: Option<String>,
}

#[derive(Clone, gotham_derive::StateData)]
pub struct State {
    pub db: Arc<Mutex<Database>>,
    pub import_queue: ImportQueue,
    pub runtime_settings: Arc<Mutex<RuntimeSettings>>,
}
impl State {
    pub fn from_db(db: Database) -> Self {
        Self {
            db: Arc::new(Mutex::new(db)),
            import_queue: ImportQueue::default(),
            runtime_settings: Default::default(),
        }
    }
}

pub fn import_entry(db: &Database, entry: &ImportQueueEntry) {
    log::info!("importing {:?}", entry.path);

    let mut last = 0.0;

    *(entry.status.lock().unwrap()) = ImportStatus::Loading(0.0);

    let r = crate::input::file_import(db, Path::new(&entry.path), move |progress| {
        // only touch the mutex if we have a meaningful change to do (1%)
        if progress > last + 0.01 {
            *(entry.status.lock().unwrap()) = ImportStatus::Loading(progress);
            last = progress;
        }
    });

    *(entry.status.lock().unwrap()) = match r {
        Ok(_) => ImportStatus::Finished,
        Err(e) => ImportStatus::Failed(format!("{}", e)),
    };

    if entry.delete_after {
        if let Err(e) = std::fs::remove_file(&entry.path) {
            log::warn!("couldn't remove imported file {}: {}", entry.path, e);
        }
    }
}

pub fn spawn_importer(state: &State) {
    let db = state.db.clone();
    let queue = state.import_queue.clone();

    std::thread::Builder::new()
        .name("import".into())
        .spawn(move || loop {
            let entry = match queue.next() {
                Some(task) => task,
                None => {
                    std::thread::sleep(std::time::Duration::from_millis(250));
                    continue;
                }
            };
            import_entry(&db.lock().unwrap(), &entry);
        })
        .expect("failed to start importer thread");
}


#[derive(serde::Serialize)]
pub struct AppStatus {
    version: String,
    storage_location: String,
    storage_size: u64,
    import_queue: Vec<serde_json::Value>,
    attachments_url: Option<String>,
    changelog: String,
}
impl AppStatus {
    pub fn new(state: &State) -> AppStatus {
        let (db_loc, db_siz) = match state.db.try_lock() {
            Ok(db) => (format!("{:?}", db.path), db.get_size()),
            Err(_) => ("".into(), 0),
        };
        AppStatus {
            storage_location: db_loc,
            storage_size: db_siz,
            import_queue: state
                .import_queue
                .status()
                .iter()
                .map(|e| {
                    serde_json::json!({
                        "path": e.path,
                        "status": e.status.lock().unwrap().to_string(),
                        "deleted_after": e.delete_after,
                    })
                })
                .collect(),
            attachments_url: state
                .runtime_settings
                .lock()
                .unwrap()
                .attachments_url
                .clone(),
            version: env!("CARGO_PKG_VERSION").to_owned(),
            changelog: crate::common::changelog(),
        }
    }
}
