use fnv::FnvHashSet;
use std::collections::HashMap;
use tantivy::collector::{Collector, SegmentCollector};
use tantivy::{DocId, Score, SegmentOrdinal, SegmentReader, DocAddress};


pub struct InspectCollector;
impl Collector for InspectCollector {
    type Fruit = Vec<tantivy::DocAddress>;
    type Child = SegmentInspectCollector;

    fn for_segment(
        &self,
        ord: SegmentOrdinal,
        segment_reader: &SegmentReader,
    ) -> tantivy::Result<Self::Child> {
        let mut ids: Vec<tantivy::DocAddress> = Vec::new();

        for doc_id in segment_reader.doc_ids_alive() {
            ids.push(DocAddress::new(ord, doc_id));
        }

        Ok(SegmentInspectCollector { ids })
    }

    fn requires_scoring(&self) -> bool {
        false
    }

    fn merge_fruits(&self, segment_counts: Vec<Self::Fruit>) -> tantivy::Result<Self::Fruit> {
        let mut ids: Vec<tantivy::DocAddress> = Vec::new();
        let it = segment_counts
            .into_iter()
            .map(|map| map.into_iter())
            .flatten();
        for addr in it {
            ids.push(addr);
        }
        Ok(ids)
    }
}

#[derive(Default)]
pub struct SegmentInspectCollector {
    ids: Vec<tantivy::DocAddress>,
}
impl SegmentCollector for SegmentInspectCollector {
    type Fruit = Vec<tantivy::DocAddress>;
    fn collect(&mut self, _: DocId, _: Score) {}
    fn harvest(self) -> Vec<tantivy::DocAddress> {
        self.ids
    }
}


type DailyCountValues = HashMap<(u32, u64), u32>;

#[derive(Default)]
pub struct DailyCount;
impl Collector for DailyCount {
    type Fruit = DailyCountValues;
    type Child = SegmentCountCollector;

    fn for_segment(
        &self,
        _: SegmentOrdinal,
        segment_reader: &SegmentReader,
    ) -> tantivy::Result<SegmentCountCollector> {
        Ok(SegmentCountCollector {
            counts: HashMap::new(),
            timeu_reader: segment_reader.fast_fields().u64("timeu").unwrap(),
            source_type_reader: segment_reader
                .fast_fields()
                .u64("source_type")
                .unwrap(),
        })
    }

    fn requires_scoring(&self) -> bool {
        false
    }

    fn merge_fruits(
        &self,
        segment_counts: Vec<DailyCountValues>,
    ) -> tantivy::Result<DailyCountValues> {
        let mut counts: HashMap<(u32, u64), u32> = HashMap::new();
        let it = segment_counts
            .into_iter()
            .map(|map| map.into_iter())
            .flatten();
        for (key, c) in it {
            counts.entry(key).and_modify(|d| *d += c).or_insert(c);
        }
        Ok(counts)
    }
}

pub struct SegmentCountCollector {
    counts: HashMap<(u32, u64), u32>,
    timeu_reader: tantivy::fastfield::Column,
    source_type_reader: tantivy::fastfield::Column,
}

impl SegmentCollector for SegmentCountCollector {
    type Fruit = DailyCountValues;

    fn collect(&mut self, id: DocId, _: Score) {
        let t: u64 = self.timeu_reader.first(id).unwrap_or_default();
        let source: u64 = self.source_type_reader.first(id).unwrap_or_default();
        let day = (t / 1_000000_000) as u32;
        self.counts
            .entry((day, source))
            .and_modify(|c| *c += 1)
            .or_insert(1);
    }

    fn harvest(self) -> DailyCountValues {
        self.counts
    }
}

pub struct IdCollector;
impl Collector for IdCollector {
    type Fruit = FnvHashSet<u64>;
    type Child = SegmentIdCollector;

    fn for_segment(
        &self,
        _: SegmentOrdinal,
        segment_reader: &SegmentReader,
    ) -> tantivy::Result<Self::Child> {
        let mut ids: FnvHashSet<u64> = FnvHashSet::default();

        let reader = segment_reader.fast_fields().u64("id");
        if let Ok(id) = reader {
            for doc_id in segment_reader.doc_ids_alive() {
                if let Some(id) = id.first(doc_id) {
                    ids.insert(id);
                }
            }
        }

        Ok(SegmentIdCollector { ids })
    }

    fn requires_scoring(&self) -> bool {
        false
    }

    fn merge_fruits(&self, segment_counts: Vec<Self::Fruit>) -> tantivy::Result<Self::Fruit> {
        let mut ids: FnvHashSet<u64> = FnvHashSet::default();
        let it = segment_counts
            .into_iter()
            .map(|map| map.into_iter())
            .flatten();
        for n in it {
            ids.insert(n);
        }
        Ok(ids)
    }
}

#[derive(Default)]
pub struct SegmentIdCollector {
    ids: FnvHashSet<u64>,
}
impl SegmentCollector for SegmentIdCollector {
    type Fruit = FnvHashSet<u64>;
    fn collect(&mut self, _: DocId, _: Score) {}
    fn harvest(self) -> FnvHashSet<u64> {
        self.ids
    }
}
