use anyhow::{anyhow, Result};
use fnv::FnvHashSet;
use itertools::Itertools;
use ring::digest::{digest, SHA256};
use std::borrow::Cow;
use std::cell::RefCell;
use std::io::Read;
use std::path::Path;
use std::path::PathBuf;
use tantivy::collector::TopDocs;
use tantivy::query::QueryParser;
use tantivy::{DocId, TantivyDocument, Index, IndexSettings, SegmentReader};

use super::collectors::{DailyCount, IdCollector};
use super::schema::{item_type, source_type, SCHEMA};
use super::types::{DailyPoint, Sort};
use super::insert::DatabaseInserter;

pub fn attachment_key(source: &str, key: &str) -> String {
    hex::encode(digest(&SHA256, format!("{}/{}", source, key).as_bytes()))
}

/// handle to the local persistent database, and methods to
/// seach and add documents. an abstraction around the search engine
/// and file system details.
pub struct Database {
    pub path: PathBuf,
    /// structure optimized for text search and comparisons
    pub index: RefCell<Index>,
    /// file tree for blobs
    pub attachments_dir: PathBuf,
}
impl Database {
    pub fn open(path: &PathBuf, overwrite: bool) -> Result<Self> {
        let index = if path.exists() {
            log::info!("loading database in {:?}", path);

            let dir = tantivy::directory::MmapDirectory::open(path)
                .map_err(|e| anyhow!("failed to open index directory: {}", e))?;

            use tantivy::TantivyError as TE;
            let index = Index::open_or_create(dir, SCHEMA.schema.clone());
            let index = match index {
                Ok(idx) => Ok(idx),
                Err(TE::IncompatibleIndex(i)) => {
                    log::warn!("incompatible index found: {:?}", i);
                    if overwrite {
                        let dir = tantivy::directory::MmapDirectory::open(path)
                            .map_err(|e| anyhow!("failed to open index directory: {}", e))?;
                        Index::create(dir, SCHEMA.schema.clone(), IndexSettings::default())
                    } else {
                        Err(TE::IncompatibleIndex(i))
                    }
                }
                Err(TE::SchemaError(s)) => {
                    log::warn!("incompatible index schema found: {:?}", s);
                    if overwrite {
                        let dir = tantivy::directory::MmapDirectory::open(path)
                            .map_err(|e| anyhow!("failed to open index directory: {}", e))?;
                        Index::create(dir, SCHEMA.schema.clone(), IndexSettings::default())
                    } else {
                        Err(TE::SchemaError(s))
                    }
                }
                Err(e) => Err(e),
            }
            .map_err(|e| anyhow!("failed to open index: {}", e))?;

            log::debug!("opened db at {:?}", path);
            index
        } else {
            log::info!("creating database in {:?}", path);
            std::fs::create_dir_all(path)?;
            let index = Index::create_in_dir(path, SCHEMA.schema.clone())
                .map_err(|e| anyhow!("failed to create index: {}", e))?;
            log::debug!("created db at {:?}", path);
            index
        };

        Ok(Self {
            path: path.clone(),
            index: RefCell::new(index),
            attachments_dir: path.join("attachments"),
        })
    }

    pub fn inspect<F: FnMut(TantivyDocument) -> ()>(&self, mut cb: F) -> Result<()> {
        let reader = self
            .index
            .borrow()
            .reader_builder()
            .reload_policy(tantivy::ReloadPolicy::OnCommitWithDelay)
            .try_into()
            .map_err(|e| anyhow!("failed to open reader: {}", e))?;

        let searcher = reader.searcher();
        let query = tantivy::query::AllQuery;

        let ids = searcher
            .search(&query, &super::collectors::InspectCollector)
            .map_err(|e| anyhow!("failed to query ids: {}", e))?;

        log::trace!("inspect: iterating over {} ids", ids.len());
        let start = std::time::Instant::now();
        for id in ids {
            match searcher.doc(id) {
                Ok(doc) => {
                    cb(doc);
                }
                Err(e) => {
                    log::warn!("could not read doc {id:?}: {e}");
                }
            }
        }
        let stop = std::time::Instant::now();
        log::trace!("inspect: finished loop in {:.3}s", (stop - start).as_secs_f32());

        Ok(())
    }

    fn parse_query(&self, q: &str) -> Result<Box<dyn tantivy::query::Query>> {
        if q.is_empty() {
            return Ok(Box::new(tantivy::query::AllQuery));
        }

        macro_rules! enum_value {
            ($q:ident, $key:expr, $s:expr, $v:expr) => {
                $q.replace(
                    std::concat!($key, ":", $s),
                    &format!(std::concat!($key, ":{}"), $v as u64),
                )
            };
        }

        let q = enum_value!(q, "source_type", "mastodon", source_type::MASTODON);
        let q = enum_value!(q, "source_type", "twitter", source_type::TWITTER);
        let q = enum_value!(q, "item_type", "post", item_type::POST);
        let q = enum_value!(q, "item_type", "boost", item_type::BOOST);

        // quote urls
        // otherwise tantivy tries to open a file? TODO
        let r = regex::Regex::new("([a-z0-9]+://[^ ]+)").unwrap();
        let q = r.replace_all(&q, "\"$1\"");

        // add an implicit "* AND " to negative queries
        let q = if q.starts_with("-") || q.starts_with("NOT ") {
            Cow::Owned(format!("* AND {}", q))
        } else {
            q
        };

        let mut query_parser = QueryParser::for_index(
            &self.index.borrow(),
            vec![SCHEMA.url, SCHEMA.body, SCHEMA.attachments, SCHEMA.summary],
        );
        query_parser.set_conjunction_by_default();

        let query = query_parser
            .parse_query(&q)
            .map_err(|e| anyhow!("failed to parse query: {}", e))?;

        Ok(query)
    }

    pub fn search(
        &self,
        q: &str,
        offset: usize,
        limit: usize,
        sort: Sort,
    ) -> Result<(usize, Vec<TantivyDocument>)> {
        let reader = self
            .index
            .borrow()
            .reader_builder()
            .reload_policy(tantivy::ReloadPolicy::OnCommitWithDelay)
            .try_into()
            .map_err(|e| anyhow!("failed to open reader: {}", e))?;

        let searcher = reader.searcher();
        let query = self.parse_query(q)?;

        let reverse = false;
        let mut top_docs = match sort {
            Sort::Time(asc) => {
                let collector = TopDocs::with_limit(limit).and_offset(offset).custom_score(
                    move |segment_reader: &SegmentReader| {
                        let time_reader = segment_reader.fast_fields().u64("timeu").unwrap();

                        move |doc: DocId| {
                            let timeu = time_reader.first(doc).unwrap_or_default();
                            if asc {
                                99990000_000000_000_u64.checked_sub(timeu).unwrap_or(0)
                            } else {
                                timeu
                            }
                        }
                    },
                );
                searcher
                    .search(&query, &collector)
                    .map_err(|e| anyhow!("failed to query: {}", e))?
            }
            Sort::Media(asc) => {
                let collector = TopDocs::with_limit(limit).and_offset(offset).custom_score(
                    move |segment_reader: &SegmentReader| {
                        let field_reader = segment_reader
                            .fast_fields()
                            .u64("attachment_count")
                            .unwrap();
                        move |doc: DocId| {
                            let value = field_reader.first(doc).unwrap_or_default();
                            if asc {
                                9999_u64.checked_sub(value).unwrap_or(0)
                            } else {
                                value
                            }
                        }
                    },
                );
                searcher
                    .search(&query, &collector)
                    .map_err(|e| anyhow!("failed to query: {}", e))?
            }
        };

        let count = searcher
            .search(&query, &tantivy::collector::Count)
            .map_err(|e| anyhow!("failed to query total: {}", e))?;

        if reverse {
            top_docs.reverse();
        }

        let docs = top_docs
            .into_iter()
            .map(|(_score, doc_address)| searcher.doc(doc_address))
            .collect::<Result<Vec<TantivyDocument>, _>>()
            .map_err(|e| anyhow!("failed to get doc: {}", e))?;

        Ok((count, docs))
    }
    pub fn query_counts(&self, q: &str) -> Result<Vec<DailyPoint>> {
        let start = std::time::Instant::now();
        let reader = self
            .index
            .borrow()
            .reader_builder()
            .reload_policy(tantivy::ReloadPolicy::OnCommitWithDelay)
            .try_into()
            .map_err(|e| anyhow!("failed to open reader: {}", e))?;

        let searcher = reader.searcher();
        let query = self.parse_query(q)?;
        log::trace!("{:?}", query);

        let counts: Vec<DailyPoint> = searcher
            .search(&query, &DailyCount::default())
            .map_err(|e| anyhow!("failed to query: {}", e))?
            .into_iter()
            .map(|((t, src), c)| (t, src, c))
            .sorted_by(|a, b| Ord::cmp(&a.0, &b.0))
            .group_by(|(t, _src, _c)| *t)
            .into_iter()
            .map(|(t, group)| {
                let mut p = DailyPoint {
                    time: t,
                    twitter: 0,
                    mastodon: 0,
                };
                group.for_each(|(_t, src, c)| {
                    if src == source_type::TWITTER as u64 {
                        p.twitter += c
                    } else if src == source_type::MASTODON as u64 {
                        p.mastodon += c
                    }
                });
                p
            })
            .collect();

        let total: u32 = counts.iter().map(|x| x.twitter + x.mastodon).sum();
        log::debug!(
            "query_counts: {} days in {:.3}s for {} entries",
            counts.len(),
            start.elapsed().as_secs_f32(),
            total
        );
        Ok(counts)
    }

    pub fn ids(&self, source_type: &str) -> Result<FnvHashSet<u64>> {
        let reader = self
            .index
            .borrow()
            .reader_builder()
            .reload_policy(tantivy::ReloadPolicy::OnCommitWithDelay)
            .try_into()
            .map_err(|e| anyhow!("failed to open reader: {}", e))?;

        let searcher = reader.searcher();
        let query = tantivy::query::TermQuery::new(
            tantivy::Term::from_field_text(SCHEMA.source_type, source_type),
            tantivy::schema::IndexRecordOption::Basic,
        );

        let ids = searcher
            .search(&query, &IdCollector)
            .map_err(|e| anyhow!("failed to query ids: {}", e))?;

        Ok(ids)
    }

    pub fn clear(&self) -> Result<()> {
        self.index
            .replace(Index::create_in_ram(SCHEMA.schema.clone()));

        erase_db_dir(&self.path)?;

        self.index.replace(
            Index::create_in_dir(&self.path, SCHEMA.schema.clone())
                .map_err(|e| anyhow!("failed to create index: {}", e))?,
        );

        if self.attachments_dir.is_dir() {
            if let Err(e) = std::fs::remove_dir_all(&self.attachments_dir) {
                log::warn!("failed to delete attachment dir: {}", e);
            }
        }
        Ok(())
    }

    pub fn inserter(&self, source_type: &'static str) -> Result<DatabaseInserter> {
        if !self.attachments_dir.is_dir() {
            if let Err(e) = std::fs::create_dir_all(&self.attachments_dir) {
                log::warn!("failed to create attachment dir: {}", e);
            }
        }

        DatabaseInserter::new(self, source_type)
    }

    pub fn get_attachment(&self, source_type: &str, key: &str) -> Result<Vec<u8>> {
        let path = self.attachments_dir.join(attachment_key(source_type, key));
        log::trace!("  = {path:?}");

        let mut f = std::fs::File::open(path)
            .map_err(|e| anyhow::anyhow!("couldnt open attachment {}: {}", key, e))?;

        let mut buffer = Vec::new();
        f.read_to_end(&mut buffer)?;

        Ok(buffer)
    }
    pub fn get_attachment_path(&self, source_type: &str, key: &str) -> String {
        self.attachments_dir
            .join(attachment_key(source_type, key))
            .to_string_lossy()
            .into_owned()
    }

    pub fn get_size(&self) -> u64 {
        dir_size(&self.path).unwrap_or(0)
    }
}

pub fn erase_db_dir(path: &Path) -> std::io::Result<()> {
    log::info!("erasing database...");
    if path.is_dir() && path.join("meta.json").is_file() {
        std::fs::remove_dir_all(&path)?;
    }
    if !path.is_dir() {
        std::fs::create_dir_all(&path)?;
    }
    Ok(())
}

fn dir_size(path: &Path) -> std::io::Result<u64> {
    fn dir_size(mut dir: std::fs::ReadDir) -> std::io::Result<u64> {
        dir.try_fold(0, |acc, file| {
            let file = file?;
            let size = match file.metadata()? {
                data if data.is_dir() => dir_size(std::fs::read_dir(file.path())?)?,
                data => data.len(),
            };
            Ok(acc + size)
        })
    }

    dir_size(std::fs::read_dir(path)?)
}
