use anyhow::{bail, Result};
use std::collections::HashMap;
use std::io::{Seek, Write};
use std::path::PathBuf;
use zip::write::SimpleFileOptions;

use super::database::Database;
use super::schema::{source_type, SCHEMA};

pub fn list_mastodon_files(source: &str) -> Result<Vec<String>> {
    let obj: HashMap<String, serde_json::Value> = serde_json::from_str(source)?;
    let mut out: Vec<String> = Vec::new();
    let items: Vec<&serde_json::Value> = obj
        .get("attachment")
        .and_then(|v| v.as_array())
        .map(|v| v.iter().collect())
        .unwrap_or_default();
    for a in items {
        if let Some(v) = a.get("url").and_then(|v| v.as_str()) {
            let v = v.strip_prefix("/").unwrap_or(v);
            out.push(v.to_owned());
        }
    }
    Ok(out)
}

pub fn list_twitter_files(source: &str) -> Result<Vec<String>> {
    let obj: HashMap<String, serde_json::Value> = serde_json::from_str(source)?;
    let mut out: Vec<String> = Vec::new();

    let entities = obj.get("extended_entities").or(obj.get("entities"))
    .and_then(|v| v.get("media"))
    .and_then(|v| v.as_array());
    let entities = match entities {
        Some(v) => v,
        None => return Ok(out),
    };

    let id_str = obj.get("id_str").and_then(|v| v.as_str());
    let id_str = match id_str {
        Some(v) => v,
        None => bail!("could not list attachments (no id_str on post)"),
    };

    for e in entities {
        let typ = e.get("type").and_then(|v| v.as_str());

        match typ {
            Some("video") | Some("animated_gif") => {
                let variants = e.get("video_info").and_then(|e| e.get("variants")).and_then(|e| e.as_array());
                let variants = match variants {
                    Some(v) => v,
                    None => continue,
                };
                for variant in variants {
                    let url = variant.get("url").and_then(|v| v.as_str());
                    let url = match url {
                        Some(v) => v,
                        None => continue,
                    };
                    let (_, filename) = url.rsplit_once('/').unwrap_or(("", url));
                    let filename = format!("tweet_media/{id_str}-{filename}");
                    out.push(filename);
                }
            },
            Some("photo") | None => {
                let url = e.get("media_url").and_then(|v| v.as_str());
                let url = match url {
                    Some(v) => v,
                    None => continue,
                };
                let (_, filename) = url.rsplit_once('/').unwrap_or(("", url));
                let filename = format!("tweet_media/{id_str}-{filename}");
                out.push(filename);
            }
            _ => continue,
        }
    }

    Ok(out)
}

pub fn export_type<W: Write>(
    mut writer: W,
    db: &Database,
    source_type: source_type,
) -> Vec<String> {
    let mut attachments = Vec::new();

    db.inspect(|doc| {
        let st = match doc.get_first(SCHEMA.source_type) {
            Some(&tantivy::schema::OwnedValue::U64(st)) => match source_type::from_u64(st) {
                Some(st) => st,
                None => return,
            },
            _ => return,
        };
        if st != source_type {
            return;
        }

        let mut source = match doc.get_first(SCHEMA.source) {
            Some(&tantivy::schema::OwnedValue::Str(ref s)) => s.clone(),
            _ => return,
        };
        source.push_str("\n");

        writer.write_all(source.as_bytes()).expect("unhandled failure in export_type(), couldnt write to archive");
        if st == source_type::MASTODON {
            attachments.extend(list_mastodon_files(&source).unwrap().into_iter());
        }
        if st == source_type::TWITTER {
            attachments.extend(list_twitter_files(&source).unwrap().into_iter());
        }

    })
    .unwrap();

    return attachments
}

pub fn export<W: Write + Seek>(db: &Database, output: W) -> Result<()> {
    let mut archive = zip::ZipWriter::new(output);
    let mut attachments = HashMap::new();

    let root = "tumelune-export";

    archive
        .add_directory("tumelune-export", SimpleFileOptions::default())?;

    log::debug!("starting activitypub.jsonlines");
    archive
        .start_file(
            format!("{root}/activitypub.jsonlines"),
            SimpleFileOptions::default(),
        )?;
    attachments.insert("mastodon", export_type(&mut archive, db, source_type::MASTODON));

    log::debug!("starting twitter.jsonlines");
    archive
        .start_file(format!("{root}/twitter.jsonlines"), SimpleFileOptions::default())?;
    attachments.insert("twitter", export_type(&mut archive, db, source_type::TWITTER));

    archive
        .add_directory(&format!("{root}/attachments/"), SimpleFileOptions::default())?;

    log::debug!("archiving attachments");

    // save attachments
    for (typ, files) in attachments.into_iter() {
        let dir = format!("{root}/attachments/{typ}/");
        archive.add_directory(&dir, SimpleFileOptions::default())?;

        log::trace!("exporting {} attachments for {typ}", files.len());

        for file in files.iter() {
            let path = dir.clone() + file;

            // disable compression for attachments
            let ao = SimpleFileOptions::default().compression_method(zip::CompressionMethod::Stored);

            log::trace!("archiving {path}");
            match db.get_attachment(&typ, &file) {
                Ok(body) => {
                    archive.start_file(path.clone(), ao)?;
                    archive.write_all(&body)?;
                }
                Err(e) => {
                    log::warn!("failed to get attachment {file}: {e}");
                }
            }
        }
    }

    archive.finish()?;

    Ok(())

}



#[derive(structopt::StructOpt)]
pub struct ExportCommand {
    //#[structopt(short="p", long="path")]
    pub path: PathBuf,
}

pub fn cli_export(db: Database, args: ExportCommand) -> anyhow::Result<()> {
    let out = std::fs::File::create(args.path)?;
    export(&db, out)?;
    Ok(())
}