use anyhow::{anyhow, Result};
use fnv::FnvHashSet;
use std::io::Write;
use std::path::PathBuf;

use super::database::{attachment_key, Database};
use super::schema::{item_type, source_type, SCHEMA};

macro_rules! tantivy_setter {
    (text $name:ident) => {
        pub fn $name(&mut self, v: &str) -> &mut Self {
            self.tdoc.add_text(SCHEMA.$name, v);
            self
        }
    };
    (u64 $name:ident) => {
        pub fn $name(&mut self, v: u64) -> &mut Self {
            self.tdoc.add_u64(SCHEMA.$name, v);
            self
        }
    };
}

/// a new document to be sent to a DatabaseInserter and then saved/indexed
pub struct NewDocument {
    /// unique id is indexed in tantivy but also here
    identifier: u64,
    /// tantivy fields
    tdoc: tantivy::TantivyDocument,
    /// full source not put in tantivy
    source: Option<String>,
}
impl NewDocument {
    pub fn new() -> Self {
        Self {
            identifier: 0,
            tdoc: tantivy::TantivyDocument::new(),
            source: None,
        }
    }

    /// id is needed out of tantivy but also should be searchable
    pub fn id(&mut self, id: u64) -> &mut Self {
        self.identifier = id;
        self.tdoc.add_u64(SCHEMA.id, id);
        self
    }

    /// source is still stored in tantivy, but should probably be
    /// stored separately to keep the index compact
    pub fn source(&mut self, v: &str) -> &mut Self {
        self.tdoc.add_text(SCHEMA.source, v);
        self.source = Some(v.to_owned());
        self
    }

    // enum setters with a enum->u64 conversion

    pub fn source_type(&mut self, v: source_type) -> &mut Self {
        self.tdoc.add_u64(SCHEMA.source_type, v as u64);
        self
    }
    pub fn item_type(&mut self, v: item_type) -> &mut Self {
        self.tdoc.add_u64(SCHEMA.item_type, v as u64);
        self
    }

    // a bunch of plain and repetitive setters to directly store in tantivy

    tantivy_setter!(text url);
    tantivy_setter!(text body);
    tantivy_setter!(text privacy);
    tantivy_setter!(text flags);
    tantivy_setter!(text author);
    tantivy_setter!(text summary);

    tantivy_setter!(text attachments);
    tantivy_setter!(text attachment_types);
    tantivy_setter!(u64 attachment_count);

    /// sets the post creation time with a chrono type value,
    /// auto setting timef/timeu
    pub fn time_chrono(&mut self, time: chrono::DateTime<chrono::offset::Utc>) -> &mut Self {
        let timef = crate::input::common::timef(&time.naive_utc());
        let timeu = crate::input::common::timeu(&time.naive_utc());

        self.tdoc
            .add_date(SCHEMA.time, crate::input::common::chrono_to_tantivy(&time));
        self.tdoc.add_f64(SCHEMA.timef, timef);
        self.tdoc.add_u64(SCHEMA.timeu, timeu);
        self
    }
}

/// ref to a database with an open writer and methods to append data
pub struct DatabaseInserter {
    ids: FnvHashSet<u64>,
    index_writer: tantivy::IndexWriter,
    source_type: &'static str,
    attachments_dir: PathBuf,
}
impl DatabaseInserter {
    pub fn new(db: &Database, source_type: &'static str) -> Result<Self> {
        let index_writer = db
            .index
            .borrow()
            .writer(64_000_000)
            .map_err(|e| anyhow!("tantivy error: {}", e))?;
        log::trace!("fetching known identifiers for {}...", source_type);
        let ids = db.ids(source_type)?;
        Ok(Self {
            ids,
            index_writer,
            source_type,
            attachments_dir: db.attachments_dir.clone(),
        })
    }
    pub fn add_status(&mut self, doc: NewDocument) {
        if self.ids.contains(&doc.identifier) {
            return;
        }
        log::trace!("adding doc {} to tantivy index", doc.identifier);
        if let Err(e) = self.index_writer.add_document(doc.tdoc) {
            log::error!("failed to add doc {}: {e}", doc.identifier);
        }
    }
    pub fn add_attachment(&mut self, key: &str, data: &[u8]) -> Result<()> {
        let path = self
            .attachments_dir
            .join(attachment_key(self.source_type, key));

        let mut f = std::fs::File::create(path)
            .map_err(|e| anyhow::anyhow!("couldnt create attachment {}: {}", key, e))?;

        f.write_all(data)
            .map_err(|e| anyhow::anyhow!("couldnt write attachment {}: {}", key, e))?;
        log::trace!("saved {} bytes as {:?}", data.len(), key);
        Ok(())
    }

    pub fn commit(&mut self) -> Result<()> {
        log::trace!("committing index");
        self.index_writer
            .commit()
            .map_err(|e| anyhow!("failed to commit: {}", e))?;

        Ok(())
    }
}
