use anyhow::{anyhow, Result};
use chrono::DateTime;
use itertools::Itertools;
use std::collections::HashSet;
use std::path::Path;

use super::common::{BoxedDocIterator, DocIterator};
use super::solvent::ArchiveWrapper;
use crate::db::{NewDocument, DatabaseInserter};
use crate::db::{item_type, source_type};

pub fn extract_privacy(row: &json::JsonValue) -> &'static str {
    let at: Option<&str> = row["attributedTo"].as_str();
    let to: Vec<&str> = row["to"].members().filter_map(|m| m.as_str()).collect();
    let cc: Vec<&str> = row["cc"].members().filter_map(|m| m.as_str()).collect();

    let public = "https://www.w3.org/ns/activitystreams#Public";
    let followers = at.map(|s| s.to_string() + "/followers");

    if to.contains(&public) {
        return "public";
    }
    if cc.contains(&public) {
        return "unlisted";
    }

    let set: std::collections::HashSet<&str> = to.iter().chain(cc.iter()).copied().collect();

    if let Some(followers) = followers {
        if set.contains(followers.as_str()) {
            return "followers";
        }
    }

    if !set.is_empty() {
        return "dm";
    }

    "private"
}
pub fn extract_time(row: &json::JsonValue) -> Result<DateTime<chrono::offset::Utc>> {
    Ok(
        DateTime::parse_from_rfc3339(row["published"].as_str().ok_or(anyhow!("bark attrTo"))?)?
            .with_timezone(&chrono::offset::Utc),
    )
}

fn media_extraction(doc: &mut NewDocument, row: &json::JsonValue) -> Result<bool> {
    let attachments_text: Vec<String> = row["attachment"]
        .members()
        .filter_map(|v| v["name"].as_str().map(|s| " ".to_owned() + s))
        .collect();

    doc.attachments(&attachments_text.join(" "));
    doc.attachment_count(row["attachment"].members().len() as u64);

    let mut media_types = HashSet::new();
    for attachment in row["attachment"].members() {
        let t = &attachment["mediaType"];
        if t.is_string() {
            media_types.insert(t.to_string());
        }
    }
    doc.attachment_types(&media_types.iter().join(" "));

    Ok(!media_types.is_empty())
}

fn common_extraction(doc: &mut NewDocument, row: &json::JsonValue, url: &str) -> Result<()> {
    doc.source_type(source_type::MASTODON);
    doc.source(&row.dump());
    doc.privacy(&extract_privacy(row));

    // url related
    doc.url(url);

    if let Some(id) = url.rsplitn(2, '/').next().and_then(|s| s.parse().ok()) {
        doc.id(id);
    }

    // time
    doc.time_chrono(extract_time(row)?);

    Ok(())
}

fn extract_post(row: &json::JsonValue) -> Result<NewDocument> {
    let url = row["url"].as_str().ok_or(anyhow!("missing post url"))?;

    let mut doc = NewDocument::new();
    doc.item_type(item_type::POST);
    common_extraction(&mut doc, row, url)?;

    let body = row["content"]
        .as_str()
        .ok_or(anyhow!("missing post content"))?;
    doc.body(&body);

    let author = row["attributedTo"]
        .as_str()
        .ok_or(anyhow!("missing attributedTo"))?;
    doc.author(&author);

    if let Some(cw) = row["summary"].as_str() {
        doc.summary(&cw);
    }

    let mut features = HashSet::new();
    if row["sensitive"].as_bool() == Some(true) {
        features.insert("sensitive");
    }
    if row["inReplyTo"].is_string() {
        features.insert("reply");
    }
    if row.has_key("oneOf") || row.has_key("anyOf") {
        features.insert("poll");
    }
    if media_extraction(&mut doc, row)? {
        features.insert("media");
    }
    features.insert("post");
    doc.flags(&features.iter().join(" "));

    Ok(doc)
}

fn extract_boost(row: &json::JsonValue) -> Result<NewDocument> {
    let post_url = match &row["object"] {
        json::JsonValue::Object(o) => o["url"]
            .as_str()
            .ok_or(anyhow!("missing boosted post url"))?,
        json::JsonValue::String(i) => i,
        _ => return Err(anyhow!("missing boost url in {}", row.dump())),
    };

    // an attempt at normalizing mastodon urls
    let post_url = post_url.replace("/users/", "/@");
    let post_url = post_url.replace("/statuses/", "/");

    let url = row["id"].as_str().ok_or(anyhow!("missing boost url"))?;

    let mut doc = NewDocument::new();
    doc.item_type(item_type::BOOST)
        .body(&format!("RT <a href=\"{u}\">{u}</a>", u = post_url))
        .flags("boost");
    common_extraction(&mut doc, row, url)?;
    Ok(doc)
}

pub fn extract_entry(row: &json::JsonValue) -> Result<NewDocument> {
    match row["type"].as_str() {
        Some("Announce") => extract_boost(row),
        Some("Create") => extract_post(&row["object"]),
        Some(t) => Err(anyhow::anyhow!("unknown type {}", t)),
        None => Err(anyhow::anyhow!("missing type")),
    }
}

pub fn open(arc: ArchiveWrapper) -> Result<BoxedDocIterator> {
    if !arc.files.contains(&Path::new("outbox.json").to_path_buf()) {
        return Err(anyhow!(
            "no outbox.json found, mastodon extractor cannot continue"
        ));
    }

    let l = arc.files.len();
    Ok(Box::new(MastodonExtractor::new(arc, l)))
}

fn import_outbox(
    dbi: &mut DatabaseInserter,
    cb: &mut dyn FnMut(f32),
    outbox: json::JsonValue,
) -> Result<()> {
    let items: Vec<&json::JsonValue> = outbox["orderedItems"].members().collect();
    let len = items.len();
    log::debug!("importing outbox.json with {len} entries");
    for (i, item) in items.iter().filter(|i| i.is_object()).enumerate() {
        match extract_entry(item) {
            Ok(doc) => {
                dbi.add_status(doc);
                cb(i as f32 / len as f32);
            }
            Err(e) => {
                log::warn!("failed to extract post #{}: {}", i, e);
                log::trace!("row: {}", item.pretty(2));
            }
        };
    }
    Ok(())
}

pub struct MastodonExtractor {
    archive: ArchiveWrapper,
    filecount: usize,
}
impl MastodonExtractor {
    pub fn new(archive: ArchiveWrapper, filecount: usize) -> Self {
        MastodonExtractor { archive, filecount }
    }
}
impl DocIterator for MastodonExtractor {
    fn source_type(&self) -> &'static str {
        "mastodon"
    }
    fn process(&mut self, dbi: &mut DatabaseInserter, cb: &mut dyn FnMut(f32)) -> Result<()> {
        let mut outbox_progress = 0f32;

        let mut decoder = self.archive.decoder()?;
        decoder.for_each(|i, path, getter| {
            log::trace!("visiting {path:?}");
            if path == Path::new("outbox.json") {
                log::debug!("loading outbox.json");
                let outbox = {
                    let buffer = getter.string();
                    json::parse(&buffer)?
                };
                let filecount = self.filecount;
                log::debug!("importing outbox");
                import_outbox(
                    dbi,
                    &mut |f| cb(f * 0.5 + (i as f32 / filecount as f32) * 0.5),
                    outbox,
                )?;
                outbox_progress = 1.0;
                return Ok(());
            }

            if path.starts_with("media_attachments/") {
                let data = getter.bytes();
                let key = path.to_str().unwrap();
                dbi.add_attachment(key, &data)?;
            }

            cb(outbox_progress * 0.5 + (i as f32 / self.filecount as f32) * 0.5);
            Ok(())
        })
    }
}
