use anyhow::{Result, anyhow};
use std::path::{Path, PathBuf};

use crate::db::Database;

pub mod solvent;
pub mod common;
pub mod mastodon_archive;
pub mod mastodon_api;
pub mod twitter_archive;

use common::{BoxedDocIterator, Settings, AccountImporter};

#[derive(structopt::StructOpt)]
pub struct ImportCommand {
    //#[structopt(short="p", long="path")]
    pub paths: Vec<PathBuf>,

    #[structopt(long = "recount", help = "recreate daily counts")]
    pub recount: bool,
}


pub fn file_import<F: FnMut(f32)>(db: &Database, path: &Path, mut status_callback: F) -> Result<()> {
    let path = PathBuf::from(path);
    log::debug!("opening archive {:?}", path);

    let arc = solvent::ArchiveWrapper::new(&path)?;

    let mut it: BoxedDocIterator = if arc.has_file("outbox.json") {
        log::debug!("archive has outbox.json; to mastodon reader");
        mastodon_archive::open(arc)?
    } else {
        log::debug!("archive has no outbox.json; to twitter reader");
        twitter_archive::open(&path)?
    };

    let mut inserter = db.inserter(it.source_type())?;
    it.process(&mut inserter, &mut status_callback)?;
    inserter.commit()?;

    Ok(())
}

pub fn api_import<F: FnMut(f32)>(db: &Database, settings: Settings, mut status_callback: F) -> Result<()> {
    let account_type = settings.get("type").ok_or(anyhow!("account type missing"))?; 

    let mut it: BoxedDocIterator = if account_type == "mastodon" {
        mastodon_api::MastodonAPI::pull(settings)?
    } else {
        return Err(anyhow!("unknown account type"));
    };

    let mut inserter = db.inserter(it.source_type())?;
    it.process(&mut inserter, &mut status_callback)?;
    inserter.commit()?;

    Ok(())
}
