///! harden the glue between all those archive formats
use anyhow::bail;
use anyhow::Result;
use flate2::read::GzDecoder;
use std::fs::File;
use std::io::Read;
use std::path::Path;
use std::path::PathBuf;

// that thing needs to be made generic over a dyn Read
pub enum FileGetter<'a> {
    TarGz(tar::Entry<'a, GzDecoder<File>>),
    Zip(zip::read::ZipFile<'a>),
}
impl<'a> FileGetter<'a> {
    pub fn bytes(self) -> Vec<u8> {
        match self {
            Self::TarGz(mut e) => {
                let mut data = Vec::new();
                if let Err(e) = e.read_to_end(&mut data) {
                    log::warn!("error reading: {e}");
                }
                data
            }
            Self::Zip(mut e) => {
                let mut data = Vec::new();
                if let Err(e) = e.read_to_end(&mut data) {
                    log::warn!("error reading: {e}");
                }
                data
            }
        }
    }
    pub fn string(self) -> String {
        match self {
            Self::TarGz(mut e) => {
                let mut buffer = String::new();
                if let Err(e) = e.read_to_string(&mut buffer) {
                    log::warn!("error reading: {e}");
                }
                buffer
            }
            Self::Zip(mut e) => {
                let mut buffer = String::new();
                if let Err(e) = e.read_to_string(&mut buffer) {
                    log::warn!("error reading: {e}");
                }
                buffer
            }
        }
    }
}

pub enum ArchiveDecoder {
    TarGz(tar::Archive<GzDecoder<File>>),
    Zip(zip::ZipArchive<File>),
}
impl ArchiveDecoder {
    pub fn from_path(path: &PathBuf) -> Result<Self> {
        let filename = path.file_name().map(|v| v.to_string_lossy().into_owned()).unwrap_or_default();
        if filename.ends_with(".tar.gz") {
            Ok(Self::TarGz(tar::Archive::new(GzDecoder::new(File::open(
                path,
            )?))))
        } else if filename.ends_with(".zip") {
            Ok(Self::Zip(zip::ZipArchive::new(std::fs::File::open(path)?)?))
        } else {
            bail!("could not detect archive type {path:?}");
        }
    }
    pub fn file_list(self) -> Result<Vec<PathBuf>> {
        Ok(match self {
            Self::TarGz(mut archive) => archive
                .entries()?
                .filter_map(|e| e.ok())
                .filter_map(|e| e.path().ok().map(|p| p.to_path_buf()))
                .collect(),
            Self::Zip(archive) => archive
                .file_names()
                .map(ToOwned::to_owned)
                .map(|s| PathBuf::from(s))
                .collect(),
        })
    }

    /// iterate over files, calling a (n, path, f) for each, with f() reading the file contents
    pub fn for_each<F: FnMut(usize, &Path, FileGetter) -> Result<()>>(
        &mut self,
        mut f: F,
    ) -> Result<()> {
        log::trace!("going into the foearchsasdafsdfS");
        Ok(match self {
            &mut Self::TarGz(ref mut archive) => {
                for (i, entry) in archive.entries()?.enumerate() {
                    let entry = match entry {
                        Ok(p) => p,
                        Err(e) => {
                            log::warn!("failed to read archive entry {}", e);
                            continue;
                        }
                    };
                    let path = match entry.path().map(|p| p.to_path_buf()) {
                        Ok(p) => p,
                        Err(e) => {
                            log::warn!("failed to read archive entry path {}", e);
                            continue;
                        }
                    };
                    f(i, &path, FileGetter::TarGz(entry))?;
                }
            }
            Self::Zip(ref mut archive) => {
                let keys: Vec<String> = archive
                    .file_names()
                    .map(ToOwned::to_owned)
                    .collect();
                for (i, name) in keys.iter().enumerate() {
                    if let Ok(entry) = archive.by_name(name) {
                        let path = PathBuf::from(name);
                        f(i, &path, FileGetter::Zip(entry))?;
                    }
                }
            }
        })
    }
}

/// a generic tool to list or iterate different kinds of archives
pub struct ArchiveWrapper {
    path: PathBuf,
    pub files: Vec<PathBuf>,
}
impl ArchiveWrapper {
    pub fn new(path: &PathBuf) -> Result<Self> {
        log::trace!("listing files");
        let list_dec = ArchiveDecoder::from_path(path)?;
        let files: Vec<PathBuf> = list_dec.file_list()?;
        log::trace!("found {} files", files.len());
        Ok(Self {
            path: path.clone(),
            files,
        })
    }
    pub fn decoder(&self) -> Result<ArchiveDecoder> {
        ArchiveDecoder::from_path(&self.path)
    }

    pub fn has_file(&self, path: &str) -> bool {
        self.files.iter().any(|v| v.to_str() == Some(path))
    }
}
