use anyhow::anyhow;
use anyhow::Context;
use anyhow::Result;
use chrono::DateTime;
use itertools::Itertools;
use serde::{Deserialize, Serialize};
use std::collections::HashSet;
use std::fs::File;
use std::io::Cursor;
use std::io::Read;
use std::path::PathBuf;
use std::sync::atomic::{AtomicUsize, Ordering};
use zip::read::ZipFile;
use zip::ZipArchive;

use super::common::{BoxedDocIterator, DocIterator};
use crate::db::{DatabaseInserter, NewDocument};
use crate::db::{item_type, source_type};

fn read_vec<R: Read>(mut r: R) -> Result<Vec<u8>> {
    let mut buffer = Vec::new();
    r.read_to_end(&mut buffer)?;
    Ok(buffer)
}

#[derive(Debug, Serialize, Deserialize)]
struct CsvTweet {
    tweet_id: usize,
    in_reply_to_status_id: Option<usize>,
    in_reply_to_user_id: Option<usize>,
    retweeted_status_id: Option<usize>,
    retweeted_status_user_id: Option<usize>,
    timestamp: chrono::DateTime<chrono::Utc>,
    source: String,
    text: String,
    #[serde(default)]
    expanded_urls: Option<String>,
}

struct CsvExtractor {
    tweets: Vec<CsvTweet>,
    position: AtomicUsize,
}
impl CsvExtractor {
    fn new(rdr: &mut csv::Reader<zip::read::ZipFile>) -> Self {
        Self {
            tweets: rdr.deserialize().filter_map(|e| e.ok()).collect(),
            position: AtomicUsize::new(0),
        }
    }
}
impl DocIterator for CsvExtractor {
    fn source_type(&self) -> &'static str {
        "twitter"
    }
    fn process(&mut self, dbi: &mut DatabaseInserter, cb: &mut dyn FnMut(f32)) -> Result<()> {
        loop {
            let pos: usize = self.position.load(Ordering::Relaxed);

            if pos >= self.tweets.len() {
                break;
            }

            let tweet = &self.tweets[pos];
            self.position.store(pos + 1, Ordering::Relaxed);

            let mut doc = NewDocument::new();
            doc.id(tweet.tweet_id as u64);
            doc.url(&format!(
                "https://twitter.com/@Twitter/statuses/{}",
                tweet.tweet_id
            ));
            doc.body(&tweet.text);
            doc.attachments(tweet.expanded_urls.as_deref().unwrap_or(""));

            doc.source_type(source_type::TWITTER);
            doc.source(&serde_json::to_string(&tweet).unwrap());

            doc.time_chrono(tweet.timestamp);
            doc.privacy("public");

            dbi.add_status(doc);
            cb(pos as f32 / self.tweets.len() as f32);
        }
        Ok(())
    }
}

fn process_attachment<T: Read + std::io::Seek>(
    dbi: &mut DatabaseInserter,
    cb: &mut dyn FnMut(f32),
    archive: &mut zip::ZipArchive<T>,
    path: &str,
) -> Result<()> {
    log::trace!("processing {}", path);
    if path.ends_with(".zip") {
        // zips in a zip. sometimes. thank u twitter
        let buffer = read_vec(archive.by_name(path)?)?;
        let mut subarchive = zip::ZipArchive::new(Cursor::new(&buffer))?;
        let subfiles = subarchive
            .file_names()
            .map(|s| s.to_string())
            .collect::<Vec<String>>();
        for (j, path) in subfiles.iter().enumerate() {
            let data = read_vec(subarchive.by_name(&path)?)?;
            dbi.add_attachment(&format!("tweet_media/{}", path), &data)
                .with_context(|| "failed to save attachment")?;
            cb(j as f32 / subfiles.len() as f32);
        }
    } else {
        let mut file = archive.by_name(&path)?;
        let mut buffer = Vec::new();
        file.read_to_end(&mut buffer)?;

        // strip data/
        let path = path.strip_prefix("data/").unwrap_or(path);

        // strip tweet_data/ and tweets_data/
        let path = path.strip_prefix("tweet_media/").unwrap_or(path);
        let path = path.strip_prefix("tweets_media/").unwrap_or(path);

        let path = format!("tweet_media/{path}");

        dbi.add_attachment(&path, &buffer)
            .with_context(|| "failed to save attachment")?;
    }
    Ok(())
}

struct TweetExtractor {
    tweets: Vec<json::JsonValue>,
    position: AtomicUsize,
    archive: zip::ZipArchive<std::fs::File>,
}
impl TweetExtractor {
    fn new(v: Vec<json::JsonValue>, archive: zip::ZipArchive<std::fs::File>) -> Self {
        Self {
            tweets: v,
            position: AtomicUsize::new(0),
            archive,
        }
    }
}
impl DocIterator for TweetExtractor {
    fn source_type(&self) -> &'static str {
        "twitter"
    }
    fn process(&mut self, dbi: &mut DatabaseInserter, cb: &mut dyn FnMut(f32)) -> Result<()> {
        // list attachments
        let attachments: Vec<String> = self
            .archive
            .file_names()
            .filter(|s| {
                s.starts_with("tweet_media/")
                    || s.starts_with("data/tweet_media/")
                    || s.starts_with("data/tweets_media/")
            })
            .filter(|s| !s.ends_with("/")) // skip directory entries
            .map(|s| s.to_string())
            .collect();

        let tweet_count = self.tweets.len();
        let total_progress = tweet_count as f32 + attachments.len() as f32;

        // process every post
        loop {
            let pos: usize = self.position.load(Ordering::Relaxed);

            if pos >= self.tweets.len() {
                break;
            }

            let tweet = &self.tweets[pos];
            self.position.store(pos + 1, Ordering::Relaxed);

            match read_json_post(tweet) {
                Ok(p) => {
                    dbi.add_status(p);
                    cb(pos as f32 / total_progress as f32);
                }
                Err(e) => log::warn!("failed to decode post {} {:?}", e, tweet),
            }
        }

        // process every file
        for (i, path) in attachments.iter().enumerate() {
            let mut subcb = |f| cb(((tweet_count + i) as f32 + f) / total_progress as f32);
            if let Err(e) = process_attachment(dbi, &mut subcb, &mut self.archive, path) {
                log::warn!("failed to process attachment #{} {}: {}", i, path, e);
            }
            cb((tweet_count + i) as f32 / total_progress as f32);
        }

        Ok(())
    }
}

fn parse_twitter_time(s: &str) -> Result<DateTime<chrono::Utc>> {
    DateTime::parse_from_str(s, "%a %b %d %H:%M:%S %z %Y")
        .or_else(|_| DateTime::parse_from_str(s, "%Y-%m-%d %H:%M:%S %z"))
        .map(|t| t.with_timezone(&chrono::Utc))
        .map_err(Into::into)
}

fn read_json_post(tweet: &json::JsonValue) -> Result<NewDocument> {
    let id = tweet["id"]
        .as_str()
        .map(|s| s.to_owned())
        .or_else(|| tweet["id"].as_usize().map(|n| n.to_string()))
        .ok_or(anyhow!("bark post id missing"))?;

    let username = tweet["user"]["screen_name"].as_str().unwrap_or("Twitter");

    let mut doc = NewDocument::new();
    if let Ok(id_n) = id.parse::<u64>() {
        doc.id(id_n);
    }
    doc.url(&format!(
        "https://twitter.com/@{}/statuses/{}",
        username, id
    ));
    doc.author(&username);

    let text = tweet["full_text"]
        .as_str()
        .or_else(|| tweet["text"].as_str())
        .unwrap_or("");

    if tweet["retweeted_status"].is_object() {
        doc.summary(&text);
        doc.item_type(item_type::BOOST);
        doc.body(&tweet["retweeted_status"]["text"].as_str().unwrap_or(""));
    } else if text.starts_with("RT @") {
        doc.item_type(item_type::BOOST);
        doc.body(&text);
    } else {
        doc.item_type(item_type::POST);
        doc.body(&text);
    }

    doc.source_type(source_type::TWITTER);
    doc.source(&json::stringify(tweet.clone()));

    let time_str = tweet["created_at"]
        .as_str()
        .ok_or(anyhow!("bark post time missing"))?;
    let time = parse_twitter_time(time_str)
        .with_context(|| format!("failed to parse time {}", time_str))?;
    doc.time_chrono(time);
    doc.privacy("public");

    let entities = if tweet["extended_entities"]["media"].is_array() {
        &tweet["extended_entities"]["media"]
    } else if tweet["entities"]["media"].is_array() {
        &tweet["entities"]["media"]
    } else {
        &json::JsonValue::Null
    };

    let mut media_types = HashSet::new();
    if entities.is_array() {
        doc.attachment_count(entities.len() as u64);
        for e in entities.members() {
            if e["type"] == "video" || e["type"] == "animated_gif" {
                media_types.insert("video");
            } else if e["type"] == "photo" || e["type"].is_null() {
                media_types.insert("image");
            } else {
                log::debug!("in {}", e.dump());
                log::warn!("unknown media type {:?}", e["type"]);
            }
        }
    }
    doc.attachment_types(&media_types.iter().join(" "));

    let mut features = HashSet::new();
    if tweet["in_reply_to_status_id"].is_number() {
        features.insert("reply");
    }
    if !media_types.is_empty() {
        features.insert("media");
    }
    doc.flags(&features.iter().join(" "));

    Ok(doc)
}

fn load_js_array<R: Read>(mut js: R) -> Result<Vec<json::JsonValue>> {
    let mut content = String::new();
    js.read_to_string(&mut content)?;
    let content = content.replace("\\/", "/");
    let array_start = content
        .find('[')
        .ok_or(anyhow!("couldnt find start of array"))?;
    let array = json::parse(&content[array_start..])?;
    Ok(array.members().cloned().collect())
}

pub fn open(path: &PathBuf) -> Result<BoxedDocIterator> {
    let mut zip = zip::ZipArchive::new(std::fs::File::open(path)?)?;

    // old archives have a set of js files for tweets and a csv
    if zip.by_name("data/js/tweet_index.js").is_ok() {
        let index = zip
            .by_name("data/js/tweet_index.js")
            .map_err(Into::into)
            .and_then(load_js_array)
            .with_context(|| "failed to load tweet index")?;

        let mut accumulator = Vec::new();
        for part in index {
            let path = part["file_name"]
                .as_str()
                .ok_or(anyhow!("failed to get part file name"))?;
            let part_tweets = zip
                .by_name(path)
                .map_err(Into::into)
                .and_then(load_js_array)
                .with_context(|| format!("failed to load part {}", path))?;
            accumulator.extend(part_tweets);
        }

        return Ok(Box::new(TweetExtractor::new(accumulator, zip)));
    }

    fn first_tweets_js<'a>(zip: &'a mut ZipArchive<File>) -> Option<ZipFile<'a>> {
        let file_list: Vec<&str> = zip.file_names().collect();
        if file_list.contains(&"data/tweet.js") {
            return zip.by_name("data/tweet.js").ok();
        }
        if file_list.contains(&"data/tweets.js") {
            return zip.by_name("data/tweets.js").ok();
        }
        return None;
    }

    // one twitter data export format, with a wrapper object
    if let Some(content) = first_tweets_js(&mut zip).and_then(|mut js| {
        let mut content = String::new();
        js.read_to_string(&mut content).ok()?;
        Some(content)
    }) {
        let array_start = content
            .find('[')
            .ok_or(anyhow!("couldnt find start of array"))?;
        let array = json::parse(&content[array_start..])?;
        let array: Vec<json::JsonValue> = array.members().map(|e| e["tweet"].clone()).collect();
        return Ok(Box::new(TweetExtractor::new(array, zip)));
    }

    // latest that i saw, with one data folder and a .html at the root
    // those do not have a {"tweet": } wrapper
    if let Ok(content) = zip.by_name("tweet.js").and_then(|mut js| {
        let mut content = String::new();
        js.read_to_string(&mut content)?;
        Ok(content)
    }) {
        let array_start = content
            .find('[')
            .ok_or(anyhow!("couldnt find start of array"))?;
        let array = json::parse(&content[array_start..])?;
        let array: Vec<json::JsonValue> = array.members().cloned().collect();
        return Ok(Box::new(TweetExtractor::new(array, zip)));
    }

    // the Classic twitter archive as a csv
    // if for some reason you only have the csv
    // the csv is encoded like shit so it's at the end to prefer a js version
    if let Ok(csv) = zip.by_name("tweets.csv") {
        let mut rdr = csv::ReaderBuilder::new().flexible(true).from_reader(csv);
        return Ok(Box::new(CsvExtractor::new(&mut rdr)));
    }

    Err(anyhow!("could not determine twitter archive type/layout"))
}
