use anyhow::{Context, Result};
use indicatif::ProgressBar;
use std::path::PathBuf;
use tantivy::schema::Value;

mod db;
mod common;
mod input;
mod web;

#[derive(structopt::StructOpt)]
struct SearchCommand {
    query: String,
}

fn default_prefix() -> PathBuf {
    dirs::executable_dir().unwrap_or_else(|| PathBuf::from("./"))
}
#[derive(structopt::StructOpt)]
struct InstallCommand {
    #[structopt(short = "p", long = "prefix")]
    prefix: Option<PathBuf>,
}

#[derive(structopt::StructOpt)]
enum Command {
    /// import new archives to index
    #[structopt(name = "import")]
    Import(input::ImportCommand),
    /// search in stored index
    #[structopt(name = "search")]
    Search(SearchCommand),
    /// start a server daemon and optionally a browser
    #[structopt(name = "serve")]
    Serve(web::ServeCommand),
    /// (for linux) copies the binary somwhere standard, with .desktop and icon files.
    #[structopt(name = "install")]
    Install(InstallCommand),
    #[structopt(name = "export")]
    Export(db::export::ExportCommand),
}

/// first ~/.config/tumelune/db then ./tumelune_db
fn default_db_path() -> std::path::PathBuf {
    // try the old config_dir if it exists
    if let Some(mut cd) = dirs::config_dir() {
        cd.push("tumelune");
        cd.push("db");

        match cd.try_exists() {
            Ok(true) => {
                log::debug!("{cd:?} found, using as the db dir");
                return cd;
            }
            Err(_) | Ok(false) => {}
        }
    }

    // otherwise use data_local_dir()
    if let Some(mut cd) = dirs::data_local_dir() {
        cd.push("tumelune");
        cd.push("db");
        log::debug!("using local data dir {cd:?}");
        return cd;
    }

    // if none are available, try the current directory
    log::debug!("no XDG directory available, using current directory");
    "./tumelune_db/".into()
}

const ABOUT: &str = "
a Thing to search in archived Posts

# import a mastodon archive
  ./tumelune import --recreate -t mastodon archive.tar.gz
# start the web ui (it's nice)
  ./tumelune serve --listen 127.0.0.1:29804 --open
  or just ./tumelune
# cli search (it's not great but it exists)
  ./tumelune search cats
";

#[derive(structopt::StructOpt)]
#[structopt(about=ABOUT)]
struct Args {
    #[structopt(subcommand)]
    cmd: Option<Command>,

    /// Path to the database
    #[structopt(short = "d", long = "db-path", env = "TL_DB_PATH")]
    db_path: Option<PathBuf>,

    #[structopt(short = "l", long = "log-level", default_value = "info")]
    log_level: String,

    #[structopt(long = "recreate", help = "delete existing database")]
    pub recreate: bool,
}

const DEFAULT_LOG: &str =
    "hyper=info,tantivy=warn,gotham=info,multipart=info,sled=info,mio=info,tracing=info";

fn prompt(text: &str) -> bool {
    loop {
        let mut buf = String::new();
        println!("{text}");
        print!("> ");
        let _ = std::io::stdin().read_line(&mut buf);

        if buf.trim().to_lowercase().as_str() == "y" {
            return true;
        }
        return false;
    }
}

fn install(c: InstallCommand) -> Result<()> {
    let prefix = c.prefix.unwrap_or_else(default_prefix);
    log::info!("installing to {prefix:?}");

    let bin_name = "tumelune";

    // directories
    let bin_dir = dirs::executable_dir().context("pick an executable dir")?;
    let icon_dir = dirs::data_dir().context("pick a data dir")?.join("icons");
    let apps_dir = dirs::data_dir()
        .context("pick a data dir")?
        .join("applications");

    // file paths
    let current_path = std::env::current_exe().context("current_exe()")?;
    let bin_path = bin_dir.join(bin_name);
    let icon_path = icon_dir.join("tumelune.png");

    // start copying

    std::fs::copy(&current_path, &bin_path)
        .with_context(|| format!("copy executable to {bin_path:?}"))?;

    std::fs::write(
        &icon_path,
        web::StaticFiles::get("favicon.png").context("")?.data,
    )
    .with_context(|| format!("copy icon to {icon_path:?}"))?;

    // all the following mess is for the .desktop file
    let desktop_file = format!(
        r"
[Desktop Entry]
Type=Application
Version=1.0
Name=tumelune
Comment=twitter and activitypub post browser and search engine
Exec={}
Icon={}
Terminal=true
",
        bin_path.to_string_lossy(),
        icon_path.to_string_lossy()
    );

    let desktop_file_path = apps_dir.join("tumelune.desktop");
    std::fs::write(desktop_file_path, &desktop_file).context("writing desktop file")?;

    let udd = which::which("update-desktop-database");
    match udd {
        Ok(udd) => {
            let mut p = std::process::Command::new(udd)
                .arg(apps_dir)
                .stdin(std::process::Stdio::piped())
                .spawn()
                .context("call update-desktop-database")?;
            p.wait().context("update-desktop-database")?;
        }
        Err(e) => {
            log::warn!("update-desktop-database not found: {e}");
        }
    }

    log::info!("installed to {bin_path:?}");

    Ok(())
}

#[paw::main]
fn main(args: Args) -> Result<()> {
    std::env::set_var(
        "RUST_LOG",
        &std::env::var("RUST_LOG")
            .unwrap_or_else(|_| format!("{},{}", DEFAULT_LOG, args.log_level)),
    );
    pretty_env_logger::init();

    log::info!("hello");
    log::debug!("using {}", tantivy::version_string());

    let mut db_path = args.db_path.clone().unwrap_or_else(default_db_path);

    if args.recreate && db_path.is_dir() {
        log::warn!("clearing db at {:?}", db_path);
        std::fs::remove_dir_all(&db_path)?;
    }

    // once the previous db is deleted, we might want to relocate to
    // the new default value
    db_path = args.db_path.clone().unwrap_or_else(default_db_path);

    let db = match db::Database::open(&db_path, true) {
        Ok(db) => db,
        Err(e) => {
            log::error!("failed to open database in {db_path:?}: {e}");
            println!("could not open database; do you want to delete and recreate it?");
            println!("(!) it will forget all previously indexed content");
            if prompt("(type y and press enter to erase, anything else or Ctrl-C to quit)") {
                match db::database::erase_db_dir(&db_path) {
                    Ok(_) => {}
                    Err(e) => {
                        log::warn!("failed to clear existing db: {e}");
                    }
                }
            }

            // reconsider location again if the previous one is corrupted
            db_path = args.db_path.unwrap_or_else(default_db_path);

            // try to open/create it again
            match db::Database::open(&db_path, true) {
                Ok(db) => db,
                Err(e) => {
                    log::error!("failed to initialize db: {e}");
                    anyhow::bail!("no database could be opened");
                }
            }
        }
    };

    match args.cmd {
        Some(Command::Import(i)) => {
            for p in &i.paths {
                log::info!("importing {:?}", p);
                let pb = ProgressBar::new(100);
                input::file_import(&db, &p, |progress| {
                    let p = (progress * 100.0).round() as u64;
                    if pb.position() < p {
                        pb.inc(p - pb.position());
                    }
                })
                .unwrap_or_else(|e| log::warn!("{:?}: import failed: {}", p, e));
            }
        }
        Some(Command::Export(args)) => {
            crate::db::export::cli_export(db, args)?;
        }
        Some(Command::Search(a)) => {
            log::info!("searching for {:?}", a.query);
            match db.search(&a.query, 0, 10000, db::Sort::Time(false)) {
                Ok((_n, docs)) => {
                    for doc in docs {
                        let source = doc
                            .get_first(crate::db::SCHEMA.source)
                            .and_then(|v| v.as_str())
                            .unwrap_or("");
                        println!("{source}");
                    }
                }
                Err(e) => log::warn!("search failed: {}", e),
            };
        }
        Some(Command::Install(c)) => {
            install(c)?;
        }
        None => {
            web::spawn_server(
                db,
                web::ServeCommand {
                    listen: None,
                    open: true,
                },
            )?;
        }
        Some(Command::Serve(a)) => {
            web::spawn_server(db, a)?;
        }
    }

    Ok(())
}
