use anyhow::bail;
use anyhow::{anyhow, Context, Result};
use bytes::BufMut;
use futures::prelude::*;
use gotham::handler::HandlerError;
use gotham::hyper::http::{header, HeaderMap, HeaderValue, StatusCode, Uri};
use gotham::hyper::{body, Body};
use gotham::router::builder::*;
use gotham::state::{FromState, State};
use itertools::Itertools;
use regex::Regex;
use rust_embed::RustEmbed;
use serde::Deserialize;
use std::borrow::Cow;
use std::fs::File;
use std::io::{Read, Seek};
use std::net::{SocketAddr, ToSocketAddrs};
use std::path::Path;
use tantivy::TantivyDocument;
use tokio::io::AsyncRead;

use crate::common::AppStatus;
use crate::db::{Database, Sort, SCHEMA};

#[derive(RustEmbed)]
#[folder = "static/"]
pub struct StaticFiles;

#[derive(structopt::StructOpt, Default, Clone)]
pub struct ServeCommand {
    /// Address to listen on.
    #[structopt(short = "l", long = "listen")]
    pub listen: Option<String>,

    /// Attempt to open with firefox or the default browser
    #[structopt(short = "o", long = "open")]
    pub open: bool,
}

#[derive(Clone, gotham_derive::StateData)]
struct WebState(crate::common::State);

use gotham_derive::{StateData, StaticResponseExtender};

#[derive(Deserialize, StateData, StaticResponseExtender)]
struct SearchQSE {
    q: String,
    #[serde(default)]
    offset: Option<u32>,
    #[serde(default)]
    max: u32,
    #[serde(default)]
    sort: String,
}

#[derive(Deserialize, StateData, StaticResponseExtender)]
struct CountQSE {
    q: String,
}

#[derive(Deserialize, StateData, StaticResponseExtender)]
struct AttachmentQSE {
    media_type: Option<String>,
}

#[derive(Deserialize, StateData, StaticResponseExtender)]
struct AttachmentPath {
    #[serde(default)]
    source_type: String,
    #[serde(rename = "*")]
    key: Vec<String>,
}

fn render_error(e: anyhow::Error) -> (mime::Mime, String) {
    let root = json::object! {
        error: format!("{}", e),
    };
    (mime::TEXT_JAVASCRIPT, json::stringify_pretty(root, 2))
}

fn clear_path(input: &str) -> String {
    let unquoter = Regex::new(r#"((^["'` \r\n]*)(file://)?|(["'` \r\n]*$))"#).unwrap();
    let s = unquoter.replace_all(input, "");
    let s = match input.contains("://") {
        true => match percent_encoding::percent_decode_str(s.as_ref()).decode_utf8() {
            Ok(dec) => dec,
            Err(_) => s,
        },
        false => s,
    };
    s.into_owned()
}
#[test]
fn test_clear_path() {
    assert_eq!(clear_path(r#"aaa"#), "aaa");
    assert_eq!(clear_path(r#" 'aaa' "#), "aaa");
    assert_eq!(clear_path(r#" 'aaa '"#), "aaa");
    assert_eq!(clear_path(r#" "aaa""#), "aaa");
    assert_eq!(clear_path(r#"'aaa` "#), "aaa");
    assert_eq!(clear_path("aaa\n"), "aaa");
    assert_eq!(clear_path("file:///te%20st"), "/te st");
    assert_eq!(clear_path("'file:///test'"), "/test");
}

async fn import_from_form(mut state: State) -> gotham::handler::HandlerResult {
    const BOUNDARY: &str = "boundary=";
    let header_map = HeaderMap::take_from(&mut state);
    let boundary = header_map
        .get(header::CONTENT_TYPE)
        .and_then(|ct| {
            let ct = ct.to_str().ok()?;
            let idx = ct.find(BOUNDARY)?;
            Some(ct[idx + BOUNDARY.len()..].to_string())
        })
        .unwrap();

    let body = body::to_bytes(Body::take_from(&mut state))
        .map_err(|e| anyhow::anyhow!(e).into())
        .await;
    let body = match body {
        Ok(v) => v,
        Err(e) => return Err((state, e)),
    };

    let mut m = multipart::server::Multipart::with_body(std::io::Cursor::new(body), boundary);
    while let Ok(Some(e)) = m.read_entry() {
        if e.headers.name.as_ref() != "file" {
            continue;
        }

        let mut data = e.data;

        let id = uuid::Uuid::new_v4();
        let filename = e.headers.filename.unwrap_or_else(|| format!("{}", id));
        let data_path = if let Some(mut cd) = dirs::config_dir() {
            cd.push("tumelune");
            cd.push("uploads");
            if let Err(e) = std::fs::create_dir_all(&cd) {
                log::warn!("failed to create directory {:?}: {:?}", cd, e);
            }
            cd.push(filename);
            cd
        } else {
            let mut pb = std::path::PathBuf::from("./");
            pb.push(filename);
            pb
        };
        let f = match std::fs::File::create(&data_path) {
            Ok(f) => f,
            Err(e) => {
                log::error!("failed to create file {:?}: {:?}", data_path, e);
                continue;
            }
        };
        let res = data.save().size_limit(Some(512 * 1000000000)).write_to(f);
        match res {
            multipart::server::SaveResult::Full(s) => {
                log::info!("uploaded {:.1}MB to {:?}", s as f32 / 1e6, data_path);
            }
            multipart::server::SaveResult::Partial(_p, reason) => {
                log::error!("failed to save the full file {:?}: {:?}", data_path, reason);
                continue;
            }
            multipart::server::SaveResult::Error(e) => {
                log::error!("failed to save the file {:?}: {:?}", data_path, e);
                continue;
            }
        }

        log::trace!("enqueuing {:?} for import", data_path);
        let ws = WebState::borrow_from(&state);
        if let Err(e) =
            ws.0.import_queue
                .push(data_path.to_string_lossy().into_owned(), true)
        {
            return Err((state, HandlerError::from(e)));
        }
    }

    ok_response(state)
}

pub fn guess_mime(path: &str) -> mime::Mime {
    match Path::new(path).extension().and_then(|oss| oss.to_str()) {
        Some("jpg") | Some("jpeg") => mime::IMAGE_JPEG,
        Some("png") => mime::IMAGE_PNG,
        Some("gif") => mime::IMAGE_GIF,
        Some("css") => mime::TEXT_CSS,
        Some("txt") => mime::TEXT_PLAIN,
        Some("js") => mime::TEXT_JAVASCRIPT,
        Some("ttf") => ttf(),
        Some("woff") => mime::FONT_WOFF,
        Some("woff2") => mime::FONT_WOFF2,
        _ => mime::APPLICATION_OCTET_STREAM,
    }
}

fn map_datetime(t: tantivy::DateTime) -> String {
    let t = t.into_offset(tantivy::time::UtcOffset::UTC);
    let format = tantivy::time::format_description::well_known::Rfc3339;
    t.format(&format)
        .unwrap_or_else(|_| "[failed to format time]".into())
}

fn map_value(tv: Option<&tantivy::schema::OwnedValue>) -> json::JsonValue {
    match tv {
        Some(tantivy::schema::OwnedValue::Str(s)) => json::JsonValue::String(s.to_owned()),
        Some(tantivy::schema::OwnedValue::PreTokStr(s)) => {
            json::JsonValue::String(s.text.to_owned())
        }
        Some(tantivy::schema::OwnedValue::U64(n)) => json::JsonValue::Number((*n).into()),
        Some(tantivy::schema::OwnedValue::I64(n)) => json::JsonValue::Number((*n).into()),
        Some(tantivy::schema::OwnedValue::F64(n)) => json::JsonValue::Number((*n).into()),
        Some(tantivy::schema::OwnedValue::Date(d)) => json::JsonValue::String(map_datetime(*d)),
        None => json::Null,
        _ => json::Null,
    }
}
fn map_value_json(tv: Option<&tantivy::schema::OwnedValue>) -> json::JsonValue {
    let s = match tv {
        Some(tantivy::schema::OwnedValue::Str(s)) => s.to_owned(),
        Some(tantivy::schema::OwnedValue::PreTokStr(s)) => s.text.to_owned(),
        _ => return json::Null,
    };
    match json::parse(&s) {
        Ok(obj) => obj,
        Err(e) => {
            log::warn!("failed to parse source: {} in {:?}", e, s);
            json::JsonValue::String(s)
        }
    }
}
fn map_source_type(tv: Option<&tantivy::schema::OwnedValue>) -> json::JsonValue {
    use crate::db::schema::source_type;
    use json::JsonValue;
    match tv {
        Some(tantivy::schema::OwnedValue::Str(s)) => json::JsonValue::String(s.to_owned()),
        Some(tantivy::schema::OwnedValue::PreTokStr(s)) => {
            json::JsonValue::String(s.text.to_owned())
        }
        Some(tantivy::schema::OwnedValue::U64(n)) => match source_type::from_u64(*n) {
            Some(source_type::MASTODON) => JsonValue::String("mastodon".into()),
            Some(source_type::TWITTER) => JsonValue::String("twitter".into()),
            _ => JsonValue::Null,
        },
        None => json::Null,
        _ => json::Null,
    }
}

async fn import(state: State) -> gotham::handler::HandlerResult {
    let ct = HeaderMap::borrow_from(&state)
        .get(header::CONTENT_TYPE)
        .and_then(|h| h.to_str().ok())
        .and_then(|s| s.parse::<mime::Mime>().ok());

    match ct {
        Some(s) if s.essence_str() == "text/plain" => import_from_path(state).await,
        Some(s) if s.essence_str() == "multipart/form-data" => import_from_form(state).await,
        Some(t) => Err((
            state,
            HandlerError::from(anyhow!("unsupported content type {}", t)),
        )),
        None => Err((state, HandlerError::from(anyhow!("no content type given")))),
    }
}

fn clear(state: State) -> (State, (mime::Mime, String)) {
    {
        let ws = WebState::borrow_from(&state);
        let db = ws.0.db.lock().unwrap();
        db.clear().unwrap();
    }
    (state, (mime::TEXT_PLAIN, "".into()))
}

fn status(state: State) -> (State, (mime::Mime, String)) {
    let json = {
        let ws = WebState::borrow_from(&state);
        let status = AppStatus::new(&ws.0);
        serde_json::to_string_pretty(&status).unwrap()
    };

    (state, (mime::TEXT_JAVASCRIPT, json))
}
fn import_queue_clear(state: State) -> (State, (mime::Mime, String)) {
    {
        let ws = WebState::borrow_from(&state);
        ws.0.import_queue.clear();
    }
    (state, (mime::TEXT_PLAIN, "".into()))
}

fn index_handler(state: State) -> (State, Result<(mime::Mime, String), HandlerError>) {
    let version_re = regex::Regex::new(r"\{\{VERSION\}\}").unwrap();
    let changelog_re = regex::Regex::new(r"\{\{CHANGELOG\}\}").unwrap();
    let version = env!("CARGO_PKG_VERSION");
    let changelog = crate::common::changelog();
    let mimetype = mime::TEXT_HTML;

    if cfg!(debug_assertions) {
        let body = std::fs::File::open("static/index.html")
            .and_then(|mut f| {
                let mut buffer = String::new();
                f.read_to_string(&mut buffer)?;
                Ok(buffer)
            })
            .map(|body| version_re.replace_all(&body, version).to_string())
            .map(|body| {
                changelog_re
                    .replace_all(&body, changelog.as_str())
                    .to_string()
            })
            .map(|body| (mimetype, body))
            .map_err(HandlerError::from);
        (state, body)
    } else {
        let index_raw = include_str!("../static/index.html");
        let index_out = version_re.replace_all(&index_raw, version);
        let index_out = changelog_re.replace_all(&index_out, changelog.as_str());
        (state, Ok((mimetype, index_out.to_string())))
    }
}

fn static_handler(
    state: State,
) -> (
    State,
    Result<(mime::Mime, Cow<'static, [u8]>), HandlerError>,
) {
    let r = {
        let uri = Uri::borrow_from(&state);
        let mut path = uri.path();
        if path.starts_with("/static/") {
            path = &path[8..];
        }
        if path.starts_with("/") {
            path = &path[1..];
        }
        log::debug!("serving static file {}", path);
        match StaticFiles::get(path) {
            Some(f) => {
                let mt = guess_mime(path);
                Ok((mt, f.data))
            }
            None => Err(HandlerError::from(anyhow::anyhow!("file not found"))),
        }
    };
    (state, r)
}

// Creates a Stream from the given file, for streaming as part of the Response.
// Stolen from gotham https://github.com/gotham-rs/gotham/blob/main/gotham/src/handler/assets/mod.rs
// Inspired by Warp https://github.com/seanmonstar/warp/blob/master/src/filters/fs.rs
// Inspired by tokio https://github.com/tokio-rs/tokio/blob/master/tokio/src/io/util/read_buf.rs
// Thanks @seanmonstar and @carllerche.
fn file_stream(
    mut f: tokio::fs::File,
    buf_size: usize,
    mut len: u64,
) -> impl TryStream<Ok = bytes::Bytes, Error = std::io::Error> + Send {
    let mut buf = bytes::BytesMut::with_capacity(buf_size);
    stream::poll_fn(move |cx| {
        if len == 0 {
            return std::task::Poll::Ready(None);
        }
        if buf.remaining_mut() < buf_size {
            buf.reserve(buf_size);
        }

        let dst = buf.chunk_mut();
        let dst = unsafe { &mut *(dst as *mut _ as *mut [std::mem::MaybeUninit<u8>]) };
        let mut read_buf = tokio::io::ReadBuf::uninit(dst);
        let read = std::pin::Pin::new(&mut f).poll_read(cx, &mut read_buf);
        std::task::ready!(read).map_err(|err| {
            log::debug!("file read error: {}", err);
            err
        })?;

        if read_buf.filled().is_empty() {
            log::debug!("file read found EOF before expected length");
            return std::task::Poll::Ready(Some(Err(std::io::Error::new(
                std::io::ErrorKind::UnexpectedEof,
                "file read found EOF before expected length",
            ))));
        }

        let n = read_buf.filled().len();
        // Safety: This is guaranteed to be the number of initialized (and read)
        // bytes due to the invariants provided by `ReadBuf::filled`.
        unsafe {
            buf.advance_mut(n);
        }
        let n = n as u64;

        let chunk = if n > len {
            let chunk = buf.split_to(len as usize);
            len = 0;
            chunk
        } else {
            len -= n;
            buf.split()
        };

        std::task::Poll::Ready(Some(Ok(chunk.freeze())))
    })
}

fn prepare_export(state: &mut State) -> Result<File> {
    let mut tmp_file = match tempfile::tempfile() {
        Ok(f) => f,
        Err(e) => {
            bail!("could not open a temporary file: {}", e);
        }
    };

    {
        let tmp_file = tmp_file.try_clone().unwrap();
        let ws = WebState::borrow_from(&state);
        let mut db = ws.0.db.lock().unwrap();
        match crate::db::export::export(&mut db, tmp_file) {
            Ok(()) => {}
            Err(e) => {
                bail!("error building archive: {}", e);
            }
        }
    }

    use std::io::SeekFrom;
    if let Err(e) = tmp_file.seek(SeekFrom::Start(0)) {
        log::warn!("failed to seek back to export archive start: {e}");
    }

    Ok(tmp_file)
}

async fn export(mut state: State) -> gotham::handler::HandlerResult {
    use gotham::hyper::header::{CONTENT_DISPOSITION, CONTENT_LENGTH, CONTENT_TYPE};
    use gotham::hyper::Response;

    log::trace!("preparing export archive...");
    let tmp_file = match prepare_export(&mut state) {
        Ok(f) => f,
        Err(e) => {
            return Err((state, HandlerError::from(e)));
        }
    };
    let flen = tmp_file.metadata().unwrap().len();
    log::debug!(
        "export tmp file ready ({:.1} MB)",
        flen as f32 / 1_000_000.0
    );

    // turn the std file into a tokio file > tokio stream for download
    let async_tmp_file = tokio::fs::File::from_std(tmp_file);
    let stream = file_stream(async_tmp_file, 1024 * 1024 * 8, flen);
    let body = Body::wrap_stream(stream.into_stream());

    let identifier = chrono::Utc::now().naive_utc().format("%Y-%m-%d--%H-%M-%S");

    let mut res = Response::new(body);
    res.headers_mut()
        .insert(CONTENT_TYPE, HeaderValue::from_static("application/zip"));
    res.headers_mut()
        .insert(CONTENT_LENGTH, HeaderValue::from(flen));
    res.headers_mut().insert(
        CONTENT_DISPOSITION,
        HeaderValue::from_str(&format!(
            "attachment; filename=\"tumelune-archive-{identifier}.zip\""
        ))
        .unwrap(),
    );

    Ok((state, res))
}

fn query(mut state: State) -> (State, (mime::Mime, String)) {
    let query_param = SearchQSE::take_from(&mut state);

    let offset = match query_param.offset {
        Some(n) if n > 0 => n,
        _ => 0,
    };
    let max = match query_param.max {
        0 => 100,
        n => n as usize,
    };
    let sort = Sort::from_str(query_param.sort.as_str());

    let results = {
        let ws = WebState::borrow_from(&state);
        let db = ws.0.db.lock().unwrap();
        db.search(&query_param.q, offset as usize, max, sort)
    };
    let (count, results): (usize, Vec<TantivyDocument>) = match results {
        Ok(r) => r,
        Err(e) => return (state, render_error(e)),
    };
    let results: json::Array = results
        .into_iter()
        .map(|doc| {
            json::object! {
                url: map_value(doc.get_first(SCHEMA.url)),
                body: map_value(doc.get_first(SCHEMA.body)),
                attachment_count: map_value(doc.get_first(SCHEMA.attachment_count)),
                attachments: map_value(doc.get_first(SCHEMA.attachments)),
                source_type: map_source_type(doc.get_first(SCHEMA.source_type)),
                source: map_value_json(doc.get_first(SCHEMA.source)),
                time: map_value(doc.get_first(SCHEMA.time)),
                timef: map_value(doc.get_first(SCHEMA.timef)),
                privacy: map_value(doc.get_first(SCHEMA.privacy)),
                summary: map_value(doc.get_first(SCHEMA.summary)),
            }
        })
        .collect();
    let root = json::object! {
        query: query_param.q,
        page_size: max,
        offset: offset,
        page: offset / max as u32,
        n_pages: count / max,
        total: count,
        results: results,
        sort: sort.as_str(),
    };

    (
        state,
        (mime::TEXT_JAVASCRIPT, json::stringify_pretty(root, 2)),
    )
}

fn query_counts(mut state: State) -> (State, (mime::Mime, String)) {
    let query_param = CountQSE::take_from(&mut state);

    let json = match {
        let ws = WebState::borrow_from(&state);
        let db = ws.0.db.lock().unwrap();

        db.query_counts(&query_param.q)
            .and_then(|c| serde_json::to_string_pretty(&c).map_err(Into::into))
    } {
        Ok(v) => v,
        Err(e) => return (state, render_error(e)),
    };

    (state, (mime::TEXT_JAVASCRIPT, json))
}

fn query_attachment(mut state: State) -> (State, gotham::hyper::Response<Body>) {
    let query_params = AttachmentQSE::take_from(&mut state);
    let path_params = AttachmentPath::take_from(&mut state);

    let key = path_params.key.iter().join("/");

    let data = {
        let ws = WebState::borrow_from(&state);
        let db = ws.0.db.lock().unwrap();
        log::debug!("fetching tree {} key {}", &path_params.source_type, &key);
        db.get_attachment(&path_params.source_type, &key)
    };

    let mime_type: mime::Mime = query_params
        .media_type
        .and_then(|s| s.parse().ok())
        .unwrap_or_else(|| guess_mime(&key));

    let r = match data {
        Ok(d) => (mime_type, d),
        Err(e) => (mime::TEXT_PLAIN, format!("error: {}", e).bytes().collect()),
    };

    let mut resp =
        gotham::helpers::http::response::create_response(&state, StatusCode::OK, r.0, r.1);

    resp.headers_mut().append(
        "Access-Control-Allow-Origin",
        gotham::hyper::header::HeaderValue::from_static("*"),
    );

    (state, resp)
}

fn ok_response(state: State) -> gotham::handler::HandlerResult {
    let resp = gotham::helpers::http::response::create_response(
        &state,
        StatusCode::OK,
        mime::TEXT_PLAIN,
        "k",
    );
    Ok((state, resp))
}

async fn import_from_path(mut state: State) -> gotham::handler::HandlerResult {
    let body = body::to_bytes(Body::take_from(&mut state))
        .map_err(|e| anyhow::anyhow!(e).into())
        .await;
    let body = match body {
        Ok(v) => v,
        Err(e) => return Err((state, e)),
    };
    let body = String::from_utf8_lossy(&body);
    for line in body.lines().map(|l| l.trim()).filter(|l| !l.is_empty()) {
        let path = clear_path(&line);
        let ws = WebState::borrow_from(&state);
        log::trace!("enqueuing {} for import", path);
        if let Err(e) = ws.0.import_queue.push(path, false) {
            return Err((state, HandlerError::from(e)));
        }
    }
    ok_response(state)
}

fn listen_addresses() -> impl Iterator<Item = std::net::SocketAddr> {
    let addresses: Vec<SocketAddr> = vec![
        "127.0.0.1:29804".parse().unwrap(),
        "127.0.0.1:29805".parse().unwrap(),
        "127.0.0.1:29806".parse().unwrap(),
        "127.0.0.1:29807".parse().unwrap(),
        "127.0.0.1:8000".parse().unwrap(),
        "127.0.0.1:0".parse().unwrap(),
        "[::1]:0".parse().unwrap(),
    ];
    addresses.into_iter()
}

pub fn serve(state: crate::common::State, args: ServeCommand) -> Result<()> {
    let ws = WebState(state);
    let state_copy = ws.0.clone();

    crate::common::spawn_importer(&ws.0);

    let middleware = gotham::middleware::state::StateMiddleware::new(ws);
    let pipeline = gotham::pipeline::single_middleware(middleware);
    let (chain, pipelines) = gotham::pipeline::single_pipeline(pipeline);

    let router = build_router(chain, pipelines, |route| {
        route.get("/").to(index_handler);
        route.get("/index.html").to(index_handler);
        route.get("/favicon.ico").to(static_handler);

        route
            .post("/search")
            .with_query_string_extractor::<SearchQSE>()
            .to(query);
        route
            .post("/count")
            .with_query_string_extractor::<CountQSE>()
            .to(query_counts);

        route
            .get("/attachment/:source_type/*")
            .with_path_extractor::<AttachmentPath>()
            .with_query_string_extractor::<AttachmentQSE>()
            .to(query_attachment);

        route.post("/import").to_async(import);
        route.get("/export").to_async(export);
        route.post("/clear").to(clear);

        route.get("/status").to(status);
        route.delete("/import_queue").to(import_queue_clear);

        route.get("*").to(static_handler);
    });

    let listen_arg = match args.listen {
        Some(l) => l
            .to_socket_addrs()
            .map_err(|e| anyhow!("failed to parse listen address: {}", e))?
            .collect(),
        None => Vec::new(),
    };

    let listens = listen_arg.into_iter().chain(listen_addresses());

    // try every address, blocking on the first one available.
    // runtime_settings.main_url will also be set to the current URL
    for listen in listens {
        log::debug!("listening on {}...", listen);
        state_copy.runtime_settings.lock().unwrap().main_url = Some(format!("http://{listen}"));
        match gotham::start(listen, router.clone()) {
            Err(e) => log::warn!("failed to start server on {listen}: {e}"),
            Ok(_) => return Ok(()),
        }
    }

    Ok(())
}

pub fn spawn_server(db: Database, args: ServeCommand) -> Result<()> {
    let state = crate::common::State::from_db(db);

    // state copies to be moved into the new thread & spawn server
    let state2 = state.clone();
    let args2 = args.clone();
    let server = std::thread::Builder::new()
        .name("server".into())
        .spawn(move || {
            serve(state2, args2).expect("failed to start the web server");
        })
        .with_context(|| "failed to start side web server thread")?;

    // then attempt to open it with a browser
    if args.open {
        loop {
            std::thread::sleep(std::time::Duration::from_secs(1));

            if server.is_finished() {
                anyhow::bail!("web server failed to start");
            }

            match state
                .runtime_settings
                .lock()
                .unwrap()
                .main_url
                .as_ref()
                .map(String::as_str)
            {
                Some(url) => {
                    let _ = open_url(url);
                    break;
                }
                None => {}
            }
        }
    }

    // if that second operation is not blocking or too short lived, we dont
    // return until the server is shut down.
    let _ = server.join();

    Ok(())
}

fn ttf() -> mime::Mime {
    "font/ttf".parse().unwrap()
}

pub fn open_url(url: &str) -> Result<()> {
    log::debug!("trying to open firefox thru webbrowser::...");
    match webbrowser::open_browser(webbrowser::Browser::Firefox, url) {
        Err(e) => log::debug!("failed to open firefox: {}", e),
        _ => return Ok(()),
    };

    log::debug!("failling back to the default browser thru webbrowser::");
    match webbrowser::open(url) {
        Err(e) => log::debug!("failed to open default browser: {}", e),
        _ => return Ok(()),
    };

    log::debug!("trying to find firefox in $PATH ourself...");
    let firefox_path = which::which("firefox");
    match firefox_path {
        Ok(path) => {
            match std::process::Command::new(&path).arg(url).spawn() {
                Ok(_) => return Ok(()),
                Err(e) => log::debug!("failed to run {:?}: {}", path, e),
            };
        }
        Err(e) => {
            log::debug!("firefox not found ({e})");
        }
    }

    anyhow::bail!("no usable browser found")
}
