function updateCssSettings() {
    const style = document.getElementById("settings_css");
    let sheet = "body {\n";

    for (const key of Object.keys(settings)) {
        const def = settings[key];
        const e = document.getElementById("settings:" + key);
        if (!e) continue;
        if (def.css_variable) {
            if (!def.test(e.value)) continue;
            const value = def.toCss(e);
            sheet += (def.css_variable + ": " + value + ";\n");
        }
    }
    sheet += "}\n";

    style.innerHTML = sheet;
}
class Settings {
    getInitial() { return 25; }
    get(key) {
        const e = document.getElementById("settings:" + key);
        return e.value;
    }
    set(key, value) {
        const e = document.getElementById("settings:" + key);
        e.value = value;
    }
    apply(key, value) {}
    restore(key, value) {}
}
class CssSettings extends Settings {
    constructor(csskey) {
        super();
        this.css_variable = csskey;
    }
    test(v) { return true; }
    apply(key, value) {
        updateCssSettings();
    }
    restore(key, value) {
    }
    getInitial() {
        const initialComputedStyle = getComputedStyle(document.body);
        let initial = initialComputedStyle.getPropertyValue(this.css_variable);
        initial = this.fromCss(initial.trim());
        return initial;
    }
}
class FontSizeSettings extends CssSettings {
    test(v) { return /^[0-9.]+$/.test(v); }
    toCss(e) { return (e.value * 1.0).toFixed(2) + 'em'; }
    fromCss(v) {
        if (v.endsWith('%')) v = v.substring(0, v.length - 1);
        if (v.endsWith('em')) v = v.substring(0, v.length - 2);
        return parseFloat(v);
    }
}
class FontFamilySettings extends CssSettings {
    toCss(e) { return `'${e.value}'`; }
    fromCss(v) {
        v = v.replaceAll(/['"]/g, '');
        return v;
    }
}
class ClassSettings extends Settings {
    constructor(prefix) {
        super()
        this.prefix = prefix;
    }
    fromClass(v) {
        return v.replace(new RegExp('^' + this.prefix + '-'), '');
    }
    toClass(v) {
        return this.prefix + '-' + v;
    }
    getInitial() {
        let initial = Array.from(document.body.classList).filter(v => v.startsWith(this.prefix + '-'))[0];
        initial = this.fromClass(initial);
        return initial;
    }
    restore(key, value) {
        const e = document.getElementById("settings:" + key);
        e.querySelectorAll('option').forEach(opt => {
            opt.selected = value == opt.value;
        });
    }
    apply(key) {
        const e = document.getElementById("settings:" + key);
        if (!e) return;

        const newClass = this.prefix + '-' + e.value;
        for (const entry of document.body.classList) {
            if (entry.startsWith(this.prefix + '-') && entry != newClass)
                document.body.classList.remove(entry);
        }
        if (!document.body.classList.contains(newClass))
            document.body.classList.add(newClass);
    }
}
const settings = {
    text_font: new FontFamilySettings("--text-font"),
    text_font_size: new FontSizeSettings("--text-font-size"),

    misc_font: new FontFamilySettings("--misc-font"),
    misc_font_size: new FontSizeSettings("--misc-font-size"),

    theme: new ClassSettings('theme'),
    page_entries: new Settings(),
}


function show_dialog(id) {
    const e = document.getElementById(id);
    if (!e) {
        console.error("dialog not found: #" + id);
        return;
    }

    // ensure all dialogs are closed
    document.querySelectorAll('#overlay > .dialog').forEach(e => {
        e.style.display = "none";
    });

    // show the one and overlay
    const overlay = document.getElementById("overlay");
    e.style.display = "block";
    overlay.style.display = "flex";
}
function close_dialogs(refresh=false) {
    // hide all dialogs
    document.querySelectorAll('#overlay > .dialog').forEach(e => {
        e.style.display = "none";
    });
    // hide overlay
    overlay.style.display = "none";

    if (refresh)
        document.getElementById("refresh").click()
}
window.addEventListener("load", function() {
    // dialogs
    const overlay = document.getElementById("overlay");
    let dialogs = [];
    function setup_dialog(div, link_id) {
        div.style.display = "none";
        if (link_id) {
            document.getElementById(link_id).onclick = function() {
                show_dialog(div.id);
                return false;
            };
        }
        // close button
        let close = document.createElement("a");
        close.innerText = "×";
        close.title = "close";
        close.classList.add("close-button");
        close.onclick = close_dialogs;
        div.prepend(close);
        dialogs.push(div);
    }
    const search_help_d = document.getElementById("search_help_d");
    const data_d = document.getElementById("data_d");
    const import_archive_d = document.getElementById("import_archive_d");
    const export_archive_d = document.getElementById("export_archive_d");
    const import_accounts_d = document.getElementById("import_accounts_d");
    const import_progress_d = document.getElementById("import_progress_d");
    const about_d = document.getElementById("about_d");
    const media_view_d = document.getElementById("media_view_d");
    const settings_d = document.getElementById("settings_d");
    setup_dialog(search_help_d, "search_help");
    setup_dialog(data_d, "import");
    setup_dialog(import_archive_d);
    setup_dialog(import_accounts_d);
    setup_dialog(import_progress_d);
    setup_dialog(export_archive_d);
    setup_dialog(about_d, "about");
    setup_dialog(media_view_d);
    setup_dialog(settings_d, "settings");
    overlay.onclick = function(ev) {
        if (ev.target != overlay) {
            return true;
        }
        close_dialogs();
        return false;
    };

    { // setup import modal
        const to_import_archive = data_d.querySelector("#to_import_archive");
        to_import_archive.onclick = function() {
            show_dialog("import_archive_d");
        };

        const to_export_archive = data_d.querySelector("#to_export_archive");
        to_export_archive.onclick = function() {
            show_dialog("export_archive_d");
        };

        const file_input = import_archive_d.querySelector("input[name=file]");
        const textarea = import_archive_d.querySelector("textarea");
        const import_button = import_archive_d.querySelector("#import_archive_send");
        const import_status = import_archive_d.querySelector("#import_archive_status");
        const import_clear = data_d.querySelector("#import_clear");
        const back_to_data = import_archive_d.querySelector("#import_archive_back");

        const clear_queue = import_progress_d.querySelector("#import_clear_queue");
        const back_to_import = import_progress_d.querySelector("#import_progress_back");
        const queue = import_progress_d.querySelector("tbody");

        let update_interval = null;
        let chosen_file = null;

        file_input.onclick = function() {
            tumelune_api.openFileDialog().then((path) => {
                console.log(path);
                if (path instanceof Array) {
                    file_input.value = path[0];
                    chosen_file = path[0];
                } else if (path instanceof File) {
                    file_input.value = path.name;
                    chosen_file = path;
                } else {
                    file_input.value = path;
                    chosen_file = path;
                }
            });
            return false;
        };
        import_clear.onclick = function() {
            const d = ": done";
            import_clear.value = import_clear.value.replace(d, "");

            tumelune_api.clearDatabase().then(() => {
                import_clear.value += d;
            });
            return false;
        };
        back_to_data.onclick = function() {
            show_dialog("data_d");
            return false;
        };

        import_button.onclick = function() {
            if (textarea.value) {
                tumelune_api.importFile({path: textarea.value});
                open_queue();
            } else if (file_input.value) {
                tumelune_api.importFile({file: chosen_file});
                open_queue();
            }
            return false;
        };

        import_status.onclick = function() {
            open_queue();
            return false;
        };

        clear_queue.onclick = function() {
            const d = ": done";
            clear_queue.value = clear_queue.value.replace(d, "");
            tumelune_api.clearImportQueue().then(() => {
                clear_queue.value += d;
            });
            return false;
        };
        back_to_import.onclick = function() {
            show_dialog("import_archive_d");
            return false;
        };

        function open_queue() {
            show_dialog("import_progress_d");
        }

        function update_status() {
            if (import_progress_d.style.display != "block" && data_d.style.display != "block") {
                return;
            }
            tumelune_api.getStatus().then((status) => {
                queue.innerHTML = "";

                document.getElementById('data_storage_size').innerText = format_bytesize(status.storage_size);
                document.getElementById('data_storage_location').innerText = status.storage_location;

                if (status.import_queue.length == 0) {
                    queue.innerHTML = "<tr class='nodata'><td colspan=2>no task in queue<br />files can take time to upload before they show here</td></tr>";
                    return;
                }

                for (const entry of status.import_queue) {
                    const row = document.createElement("tr");
                    if (entry.path.length > 80) {
                        const short = entry.path
                            .replace(/.+\/([^/]+)\/([^/]+)/, "…/$1/$2");
                        row.appendChild(tdtag(short, {long: entry.path}));
                    } else {
                        row.appendChild(tdtag(entry.path));
                    }
                    row.appendChild(tdtag(entry.status));
                    queue.appendChild(row);
                }
            });
        }

        if (update_interval === null) {
            update_interval = setInterval(update_status, 250);
        }
        update_status();
    }

    { // setup data . export

        const back_to_data = export_archive_d.querySelector("#export_archive_back");
        back_to_data.onclick = function() {
            show_dialog("data_d");
            return false;
        };
    }

    function saveSettings() {
        for (const key of Object.keys(settings)) {
            const e = document.getElementById("settings:" + key);
            window.localStorage.setItem('settings.' + key, e.value);
        }
    }

    { // setup settings modal
        const style = document.getElementById("settings_css");
        let initialValues = {};
        console.debug("config: initializing");
        for (const key of Object.keys(settings)) {
            const def = settings[key];
            const e = document.getElementById("settings:" + key);
            if (!e) continue;

            // set and collect initial values
            let initial = def.getInitial();
            initialValues[key] = initial;

            // overwrite with any localStorage saved value
            const stored = window.localStorage.getItem('settings.'+key);
            if (stored) initial = stored;

            // restore
            console.log(`config: ${key} = ${initial}`);
            e.value = initial ?? '';

            // setup change handlers
            e.onchange = () => {
                if (def.filter && !def.filter(e.value)) return;

                console.debug(`config: ${key} = ${e.value}, applying`);
                def.apply(key);

                saveSettings();
            };
        }
        // update all at once at the end, applying css involves every value
        for (const key of Object.keys(settings)) {
            const def = settings[key];
            const value = def.get(key);
            def.apply(key, value);
            def.restore(key, value);
        }
        // setup button handlers
        document.getElementById('settings_close').onclick = () => {
            close_dialogs(true);
        };
        document.getElementById('settings_reset').onclick = () => {
            for (const key of Object.keys(settings)) {
                settings[key].set(key, initialValues[key]);
                settings[key].apply(key, initialValues[key]);
                settings[key].restore(key, initialValues[key]);
            }
            saveSettings();
        };
    }
});
